"""Unit tests for the problem_report module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import base64
import contextlib
import datetime
import email
import io
import locale
import sys
import textwrap
import time
import unittest
import unittest.mock
from unittest.mock import MagicMock

try:
    import zstandard
except ImportError:
    zstandard = None  # type: ignore

import problem_report

bin_data = b"ABABABABAB\0\0\0Z\x01\x02"
GZIP_BIN_DATA = (
    b"\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x03str\x84B\x06"
    b"\x06\x86(F&\x003\x95\xd4\x0b\x10\x00\x00\x00"
)
ZSTD_BIN_DATA = (
    b"(\xb5/\xfd$\x10\x81\x00\x00ABABABABAB\x00\x00\x00Z\x01\x02\xbc\xdf\xdd\xfd"
)


class T(unittest.TestCase):  # pylint: disable=too-many-public-methods
    """Unit tests for the problem_report module."""

    def test_add_tags(self):
        """Test ProblemReport.add_tags()."""
        report = problem_report.ProblemReport()
        report.add_tags({"tag1"})
        self.assertEqual(report["Tags"], "tag1")
        report.add_tags(["tag2", "next_tag"])
        self.assertEqual(report["Tags"], "next_tag tag1 tag2")

    def test_add_tag_drop_duplicates(self):
        """Test ProblemReport.add_tags() dropping duplicates."""
        report = problem_report.ProblemReport()
        report.add_tags({"same"})
        self.assertEqual(report["Tags"], "same")
        report.add_tags(["same"])
        self.assertEqual(report["Tags"], "same")

    def test_basic_operations(self):
        """Basic creation and operation."""
        pr = problem_report.ProblemReport()
        pr["foo"] = "bar"
        pr["bar"] = " foo   bar\nbaz\n   blip  "
        pr["dash-key"] = "1"
        pr["dot.key"] = "1"
        pr["underscore_key"] = "1"
        self.assertEqual(pr["foo"], "bar")
        self.assertEqual(pr["bar"], " foo   bar\nbaz\n   blip  ")
        self.assertEqual(pr["ProblemType"], "Crash")
        locale_time = locale.getlocale(locale.LC_TIME)
        locale.setlocale(locale.LC_TIME, "C")
        try:
            self.assertTrue(time.strptime(pr["Date"]))
        finally:
            locale.setlocale(locale.LC_TIME, locale_time)
        self.assertEqual(pr["dash-key"], "1")
        self.assertEqual(pr["dot.key"], "1")
        self.assertEqual(pr["underscore_key"], "1")

    def test_ctor_arguments(self):
        """non-default constructor arguments."""
        pr = problem_report.ProblemReport("KernelCrash")
        self.assertEqual(pr["ProblemType"], "KernelCrash")
        pr = problem_report.ProblemReport(date="19801224 12:34")
        self.assertEqual(pr["Date"], "19801224 12:34")

    def test_get_timestamp(self):
        """get_timestamp() returns timestamp."""
        r = problem_report.ProblemReport()
        self.assertAlmostEqual(r.get_timestamp(), time.time(), delta=2)

        r["Date"] = "Thu Jan 9 12:00:00 2014"
        # delta is ±12 hours, as this depends on the timezone that the test is
        # run in
        self.assertAlmostEqual(r.get_timestamp(), 1389265200, delta=43200)

    def test_get_timestamp_locale_german(self):
        """get_timestamp() returns date when LC_TIME is set."""
        now = datetime.datetime.now()

        pr = problem_report.ProblemReport(date=now.strftime("%a %b %d %H:%M:%S %Y"))
        orig_ctime = locale.getlocale(locale.LC_TIME)
        try:
            locale.setlocale(locale.LC_TIME, "de_DE.UTF-8")
        except locale.Error:
            self.skipTest("Missing German locale support")
        self.assertEqual(pr.get_timestamp(), int(now.timestamp()))
        locale.setlocale(locale.LC_TIME, orig_ctime)

    def test_get_timestamp_returns_none(self):
        """get_timestamp() returns None."""
        pr = problem_report.ProblemReport()
        del pr["Date"]
        self.assertIsNone(pr.get_timestamp())

    def test_consistency_checks(self):
        """Various error conditions."""
        pr = problem_report.ProblemReport()
        self.assertRaises(ValueError, pr.__setitem__, "a b", "1")
        self.assertRaises(TypeError, pr.__setitem__, "a", 1)
        self.assertRaises(TypeError, pr.__setitem__, "a", (1,))
        self.assertRaises(TypeError, pr.__setitem__, "a", ("/tmp/nonexistent", ""))
        self.assertRaises(
            TypeError,
            pr.__setitem__,
            "a",
            ("/tmp/nonexistent", False, 0, True, "bogus"),
        )
        self.assertRaises(TypeError, pr.__setitem__, "a", ["/tmp/nonexistent"])
        self.assertRaises(KeyError, pr.__getitem__, "Nonexistent")

    def test_write(self) -> None:
        """write() and proper formatting."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["SimpleUTF8"] = "1äö2Φ3".encode("UTF-8")
        pr["SimpleUnicode"] = "1äö2Φ3"
        pr["TwoLineUnicode"] = "pi-π\nnu-η"
        pr["TwoLineUTF8"] = "pi-π\nnu-η".encode("UTF-8")
        pr["WhiteSpace"] = " foo   bar\nbaz\n  blip  \n\nafteremptyline"
        # Unicode with a non-space low ASCII character \x05 in it
        pr["UnprintableUnicode"] = b"a\xc3\xa4\x05z1\xc3\xa9".decode("UTF-8")
        out = io.BytesIO()
        pr.write(out)
        expected = (
            textwrap.dedent(
                """\
                ProblemType: Crash
                Date: now!
                Simple: bar
                SimpleUTF8: 1äö2Φ3
                SimpleUnicode: 1äö2Φ3
                TwoLineUTF8:
                 pi-π
                 nu-η
                TwoLineUnicode:
                 pi-π
                 nu-η
                UnprintableUnicode: aä\x05z1é
                WhiteSpace:
                """
            )
            + "  foo   bar\n baz\n   blip  \n \n afteremptyline\n"
        ).encode("UTF-8")
        self.assertEqual(out.getvalue(), expected)

    def test_load(self):
        """load() with various formatting."""
        report = textwrap.dedent(
            f"""\
            ProblemType: Crash
            Date: now!
            Simple: bar
            WhiteSpace:
              foo   bar
             baz
               blip{'  '}
            """
        )
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n  blip  ")

        # test last field a bit more
        report += " \n"
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n  blip  \n")

        # last field might not be \n terminated
        pr.load(
            io.BytesIO(
                textwrap.dedent(
                    """\
                    ProblemType: Crash
                    Date: now!
                    Simple: bar
                    WhiteSpace:
                     foo
                     bar"""
                ).encode()
            )
        )
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], "foo\nbar")

        report = (
            "ProblemType: Crash\n"
            "WhiteSpace:\n"
            "  foo   bar\n"
            " baz\n"
            " \n"
            "   blip  \n"
            "Last: foo\n"
        )
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n\n  blip  ")
        self.assertEqual(pr["Last"], "foo")

        report += " \n"
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n\n  blip  ")
        self.assertEqual(pr["Last"], "foo\n")

        # empty lines in values must have a leading space in coding
        invalid_spacing = io.BytesIO(
            textwrap.dedent(
                """\
                WhiteSpace:
                 first

                 second
                """
            ).encode()
        )
        pr = problem_report.ProblemReport()
        self.assertRaises(ValueError, pr.load, invalid_spacing)

        # test that load() cleans up properly
        pr.load(io.BytesIO(b"ProblemType: Crash"))
        self.assertEqual(list(pr.keys()), ["ProblemType"])

    def test_load_binary_blob(self):
        """Throw exception when binary file (e.g. core) is loaded."""
        report = problem_report.ProblemReport()
        with io.BytesIO(b"AB\xfc:CD") as report_file:
            with self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                "codec can't decode byte 0xfc in position 2",
            ):
                report.load(report_file)

    def test_load_missing_colon(self):
        """Throw exception when key-value line misses a colon as separator."""
        report = problem_report.ProblemReport()
        with io.BytesIO(b"\n") as report_file:
            with self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                r"Line '\\n' does not contain a colon",
            ):
                report.load(report_file)

    def test_load_invalid_utf8(self):
        """Throw exception when binary file is invalid UTF-8."""
        report = problem_report.ProblemReport()
        with io.BytesIO(b"\x7fELF\x02\x01\xb0j") as report_file:
            with self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                r"Line '\\x7fELF\\x02\\x01\\\\xb0j' does not contain a colon",
            ):
                report.load(report_file)

    def test_load_incorrect_padding(self):
        """Throw exception when base64 encoded data has incorrect padding."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" H4sICAAAAAAC/0NvcmVEdW1wAA==\n"
            b" 7Z0LYFPV/cdP0rQ\n"
        )
        with io.BytesIO(content) as report_file:
            with self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                "^Malformed problem report: Incorrect padding."
                " Is this a proper .crash text file\\?$",
            ):
                report.load(report_file)

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_load_zstd_compressed_data(self) -> None:
        """Test reading zstd-compressed data."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SROHQIAgoQOEqC7ASCSGCybZRrN7//Hsn7dVyActu7bbMcLaav0RC06\n"
            b" m6fCZ/N7aeOeyqxspRVn88bSx8a8opWQAwEA/8OEAhOOwLA=\n"
        )
        with io.BytesIO(content) as report_file:
            report.load(report_file)
        self.assertEqual(
            report["CoreDump"],
            "sample data that is compressed with zstd"
            " and is long enough for zstd to work.\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_reading_zstd_compressed_value(self) -> None:
        """Test reading zstd-compressed CompressedValue."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SRPHQIA8sQPFKClOoBIEkjIbmfbhiz9/1/96mYFATGiz9JHLfTZY9aV\n"
            b" a9QBRdQz0AkUjbdbVqNQorP3UpGC1NIi9X3UgygBAEyQcYs=\n"
        )
        with io.BytesIO(content) as report_file:
            report.load(report_file, binary="compressed")
        self.assertEqual(
            report["CoreDump"].get_value(),
            b"fake core dump data for testing purposes"
            b" which is long enough to be compressed\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_writing_zstd_compressed_value(self) -> None:
        """Test writing zstd-compressed CompressedValue."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(
                b"KLUv/SRCvQEAIsQMEMC3AbLcS0WaZ8rvf69jyw3ghpXr6pBr5i7HiCKroc60"
                b"3+tB4rC/4TV1osHyJUXme6IIAD6PTiM="
            )
        )
        output = io.BytesIO()
        compressed_value.write(output)
        self.assertEqual(
            output.getvalue(),
            b"Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_zstd_compressed_value_length(self) -> None:
        """Test getting length of zstd-compressed CompressedValue."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(
                b"KLUv/SRB3QEAwoMMEbDrOJRilklIQlLIJC3OvFUMIIwdaHxrV6aSHEv6rFlP"
                b"/ll/IaXfo19actUkdvm8hggBAAlhlgHbiBMF"
            )
        )
        self.assertEqual(len(compressed_value), 65)

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_len_zstd_compressed_value_nosize(self) -> None:
        """Test len() on zstd-compressed CompressedValue without a size header."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(b"KLUv/QBYEQAAe30=")
        )
        self.assertEqual(len(compressed_value), 2)

    @unittest.mock.patch("builtins.__import__")
    def test_zstandard_missing(self, import_mock: MagicMock) -> None:
        """Test reading zstd-compressed data when zstandard is missing."""
        with contextlib.suppress(KeyError):
            sys.modules.pop("zstandard")
        import_mock.side_effect = ImportError("mocked import error")

        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SROHQIAgoQOEqC7ASCSGCybZRrN7//Hsn7dVyActu7bbMcLaav0RC06\n"
            b" m6fCZ/N7aeOeyqxspRVn88bSx8a8opWQAwEA/8OEAhOOwLA=\n"
        )
        expected_message = (
            "Failed to import zstandard library: mocked import error."
            " Please install python3-zstandard."
        )
        with self.assertRaisesRegex(RuntimeError, expected_message):
            with io.BytesIO(content) as report_file:
                report.load(report_file)

    def test_write_fileobj(self):
        """Write a report with a pointer to a file-like object."""
        tempbin = io.BytesIO(bin_data)
        tempasc = io.BytesIO(b"Hello World")

        pr = problem_report.ProblemReport(date="now!")
        pr["BinFile"] = (tempbin,)
        pr["AscFile"] = (tempasc, False)
        out = io.BytesIO()
        pr.write(out)
        out.seek(0)

        pr = problem_report.ProblemReport()
        pr.load(out)
        self.assertEqual(pr["BinFile"], tempbin.getvalue())
        self.assertEqual(pr["AscFile"], tempasc.getvalue().decode())

    def test_write_empty_fileobj(self):
        """Write a report with a pointer to a file-like object with
        enforcing non-emptyness."""
        tempbin = io.BytesIO(b"")
        tempasc = io.BytesIO(b"")

        pr = problem_report.ProblemReport(date="now!")
        pr["BinFile"] = (tempbin, True, None, True)
        out = io.BytesIO()
        self.assertRaises(OSError, pr.write, out)

        pr = problem_report.ProblemReport(date="now!")
        pr["AscFile"] = (tempasc, False, None, True)
        out = io.BytesIO()
        self.assertRaises(OSError, pr.write, out)

    def test_read_file(self):
        """Read a report with binary data."""
        bin_report = textwrap.dedent(
            """\
            ProblemType: Crash
            Date: now!
            File: base64
             H4sICAAAAAAC/0ZpbGUA
             c3RyhEIGBoYoRiYAM5XUCxAAAAA=
            Foo: Bar
            """
        ).encode()

        # test with reading everything
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(bin_report))
        self.assertEqual(pr["File"], bin_data)
        self.assertEqual(pr.has_removed_fields(), False)

        # test with skipping binary data
        pr.load(io.BytesIO(bin_report), binary=False)
        self.assertIsNone(pr["File"])
        self.assertEqual(pr.has_removed_fields(), True)

        # test with keeping compressed binary data
        pr.load(io.BytesIO(bin_report), binary="compressed")
        self.assertEqual(pr["Foo"], "Bar")
        self.assertEqual(pr.has_removed_fields(), False)
        self.assertTrue(isinstance(pr["File"], problem_report.CompressedValue))
        self.assertEqual(len(pr["File"]), len(bin_data))
        self.assertEqual(pr["File"].get_value(), bin_data)
        self.assertEqual(pr["File"].name, "File")

    def test_read_file_legacy(self):
        """Read a report with binary data in legacy format without gzip
        header."""
        bin_report = textwrap.dedent(
            """\
            ProblemType: Crash
            Date: now!
            File: base64
             eJw=
             c3RyxIAMcBAFAG55BXk=
            Foo: Bar
            """
        ).encode()

        # test with reading everything
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(bin_report))
        self.assertEqual(pr["File"], b"AB" * 10 + b"\0" * 10 + b"Z")
        self.assertEqual(pr.has_removed_fields(), False)

        # test with skipping binary data
        pr.load(io.BytesIO(bin_report), binary=False)
        self.assertIsNone(pr["File"])
        self.assertEqual(pr.has_removed_fields(), True)

        # test with keeping CompressedValues
        pr.load(io.BytesIO(bin_report), binary="compressed")
        self.assertEqual(pr.has_removed_fields(), False)
        self.assertEqual(len(pr["File"]), 31)
        self.assertEqual(pr["File"].get_value(), b"AB" * 10 + b"\0" * 10 + b"Z")
        self.assertEqual(pr["File"].name, "File")
        out = io.BytesIO()
        pr["File"].write(out)
        out.seek(0)
        self.assertEqual(out.read(), b"AB" * 10 + b"\0" * 10 + b"Z")

    def test_iter(self):
        """problem_report.ProblemReport iteration."""
        pr = problem_report.ProblemReport()
        pr["foo"] = "bar"

        keys = []
        for k in pr:
            keys.append(k)
        keys.sort()
        self.assertEqual(" ".join(keys), "Date ProblemType foo")

        self.assertEqual(len([k for k in pr if k != "foo"]), 2)

    def test_modify(self):
        """reading, modifying fields, and writing back."""
        report = textwrap.dedent(
            """\
            ProblemType: Crash
            Date: now!
            Long:
             xxx
             .
             yyy
            Short: Bar
            File: base64
             H4sICAAAAAAC/0ZpbGUA
             c3RyxIAMcBAFAK/2p9MfAAAA
            """
        ).encode()

        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report), binary="compressed")

        self.assertEqual(pr["Long"], "xxx\n.\nyyy")
        self.assertEqual(
            pr["File"].get_value(), b"ABABABABABABABABABAB\0\0\0\0\0\0\0\0\0\0Z"
        )

        # write back unmodified
        out = io.BytesIO()
        pr.write(out)
        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent(
                """\
                ProblemType: Crash
                Date: now!
                Long:
                 xxx
                 .
                 yyy
                Short: Bar
                File: base64
                 H4sICAAAAAAC/0ZpbGUAc3RyxIAMcBAFAK/2p9MfAAAA
                """
            ),
        )

        pr["Short"] = "aaa\nbbb"
        pr["Long"] = "123"
        out = io.BytesIO()
        pr.write(out)
        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent(
                """\
                ProblemType: Crash
                Date: now!
                Long: 123
                Short:
                 aaa
                 bbb
                File: base64
                 H4sICAAAAAAC/0ZpbGUAc3RyxIAMcBAFAK/2p9MfAAAA
                """
            ),
        )

    def test_sorted_items(self) -> None:
        """Test ProblemReport.sorted_items()."""
        bin_report = textwrap.dedent(
            """\
            ProblemType: Crash
            Date: now!
            File: base64
             eJw=
             c3RyxIAMcBAFAG55BXk=
            _MarkForUpload: False
            Architecture: amd64
            ExecutablePath: /usr/bin/python3
            Package: python3.12-minimal
            """
        ).encode()

        report = problem_report.ProblemReport()
        report.load(io.BytesIO(bin_report), binary=False)
        self.assertEqual(
            list(report.sorted_items()),
            [
                ("ExecutablePath", "/usr/bin/python3"),
                ("Package", "python3.12-minimal"),
                ("ProblemType", "Crash"),
                ("Architecture", "amd64"),
                ("Date", "now!"),
            ],
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_write_mime_binary_values(self) -> None:
        """write_mine() for binary values (gzip and zstd compressed)."""
        report = problem_report.ProblemReport(date="now!")
        report["Data.gz"] = GZIP_BIN_DATA
        report["GzipData"] = problem_report.CompressedValue(
            compressed_value=GZIP_BIN_DATA
        )
        report["ZstdData"] = problem_report.CompressedValue(
            compressed_value=ZSTD_BIN_DATA
        )

        output = io.BytesIO()
        report.write_mime(output)
        output.seek(0)

        message = email.message_from_binary_file(output)
        remaining_parts = message.walk()

        # first part is the multipart container
        part = next(remaining_parts)
        self.assertTrue(part.is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        part = next(remaining_parts)
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "text/plain")
        self.assertEqual(part.get_content_charset(), "utf-8")
        self.assertIsNone(part.get_filename())
        self.assertEqual(
            part.get_payload(decode=True), b"ProblemType: Crash\nDate: now!\n"
        )

        # third part should be the Data.gz as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "Data.gz")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/gzip")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), GZIP_BIN_DATA)

        # fourth part should be the GzipData as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "GzipData.gz")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/gzip")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), GZIP_BIN_DATA)

        # fifth part should be the ZstdData as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "ZstdData.zst")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/zstd")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), ZSTD_BIN_DATA)

        with self.assertRaises(StopIteration):
            next(remaining_parts)

    def test_write_mime_invalid_compressed_binary(self) -> None:
        """write_mine() for invalid compressed binary values."""
        report = problem_report.ProblemReport()
        report["InvalidData.gz"] = b"\0X"

        output = io.BytesIO()
        with self.assertRaisesRegex(
            ValueError, "^Unknown compression for InvalidData.gz$"
        ):
            report.write_mime(output)

    def test_write_mime_text(self):
        """write_mime() for text values."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["SimpleUTF8"] = "1äö2Φ3".encode("UTF-8")
        pr["SimpleUnicode"] = "1äö2Φ3"
        pr["TwoLineUnicode"] = "pi-π\nnu-η\n"
        pr["TwoLineUTF8"] = "pi-π\nnu-η\n".encode("UTF-8")
        pr["SimpleLineEnd"] = "bar\n"
        pr["TwoLine"] = "first\nsecond\n"
        pr["InlineMargin"] = "first\nsecond\nthird\nfourth\nfifth\n"
        pr["Multiline.txt"] = " foo   bar\nbaz\n  blip  \nline4\nline♥5!!\nłıµ€ ⅝\n"

        # still small enough for inline text
        pr["Largeline"] = "A" * 999
        pr["LargeMultiline"] = f"{'A' * 120}\n{'B' * 90}"

        # too big for inline text, these become attachments
        pr["Hugeline"] = "A" * 10000
        pr["HugeMultiline"] = f"{'A' * 900}\n{'B' * 900}\n{'C' * 900}"
        out = io.BytesIO()
        pr.write_mime(out)
        out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 5)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        expected = textwrap.dedent(
            f"""\
            ProblemType: Crash
            Date: now!
            InlineMargin:
             first
             second
             third
             fourth
             fifth
            LargeMultiline:
             {'A' * 120}
             {'B' * 90}
            Largeline: {'A' * 999}
            Simple: bar
            SimpleLineEnd: bar
            SimpleUTF8: 1äö2Φ3
            SimpleUnicode: 1äö2Φ3
            TwoLine:
             first
             second
            TwoLineUTF8:
             pi-π
             nu-η
            TwoLineUnicode:
             pi-π
             nu-η
            """
        ).encode("UTF-8")
        self.assertEqual(parts[1].get_payload(decode=True), expected)

        # third part should be the HugeMultiline: field as attachment
        self.assertTrue(not parts[2].is_multipart())
        self.assertEqual(parts[2].get_content_type(), "text/plain")
        self.assertEqual(parts[2].get_content_charset(), "utf-8")
        self.assertEqual(parts[2].get_filename(), "HugeMultiline.txt")
        self.assertEqual(
            parts[2].get_payload(decode=True), pr["HugeMultiline"].encode()
        )

        # fourth part should be the Hugeline: field as attachment
        self.assertTrue(not parts[3].is_multipart())
        self.assertEqual(parts[3].get_content_type(), "text/plain")
        self.assertEqual(parts[3].get_content_charset(), "utf-8")
        self.assertEqual(parts[3].get_filename(), "Hugeline.txt")
        self.assertEqual(parts[3].get_payload(decode=True), pr["Hugeline"].encode())

        # fifth part should be the Multiline: field as attachment
        self.assertTrue(not parts[4].is_multipart())
        self.assertEqual(parts[4].get_content_type(), "text/plain")
        self.assertEqual(parts[4].get_content_charset(), "utf-8")
        self.assertEqual(parts[4].get_filename(), "Multiline.txt")
        expected = textwrap.dedent(
            f"""\
             foo   bar
            baz
              blip{'  '}
            line4
            line♥5!!
            łıµ€ ⅝
            """
        ).encode("UTF-8")
        self.assertEqual(parts[4].get_payload(decode=True), expected)

    def test_write_mime_extra_headers(self):
        """write_mime() with extra headers."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["TwoLine"] = "first\nsecond\n"
        out = io.BytesIO()
        pr.write_mime(out, extra_headers={"Greeting": "hello world", "Foo": "Bar"})
        out.seek(0)

        msg = email.message_from_binary_file(out)
        self.assertEqual(msg["Greeting"], "hello world")
        self.assertEqual(msg["Foo"], "Bar")
        parts = list(msg.walk())
        self.assertEqual(len(parts), 2)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertIn(b"Simple: bar", parts[1].get_payload(decode=True))

    def test_write_mime_order(self):
        """write_mime() with keys ordered."""
        pr = problem_report.ProblemReport(date="now!")
        pr["SecondText"] = "What"
        pr["FirstText"] = "Who"
        pr["FourthText"] = "Today"
        pr["ThirdText"] = "I Don't Know"
        out = io.BytesIO()
        pr.write_mime(
            out,
            priority_fields=[
                "FirstText",
                "SecondText",
                "ThirdText",
                "Unknown",
                "FourthText",
            ],
        )
        out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 2)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        self.assertEqual(
            parts[1].get_payload(decode=True),
            textwrap.dedent(
                """\
                FirstText: Who
                SecondText: What
                ThirdText: I Don't Know
                FourthText: Today
                ProblemType: Crash
                Date: now!
                """
            ).encode(),
        )

    def test_updating(self):
        """new_keys() and write() with only_new=True."""
        pr = problem_report.ProblemReport()
        self.assertEqual(pr.new_keys(), set(["ProblemType", "Date"]))
        pr.load(
            io.BytesIO(
                textwrap.dedent(
                    """\
                    ProblemType: Crash
                    Date: now!
                    Foo: bar
                    Baz: blob
                    """
                ).encode()
            )
        )

        self.assertEqual(pr.new_keys(), set())

        pr["Foo"] = "changed"
        pr["NewKey"] = "new new"
        self.assertEqual(pr.new_keys(), set(["NewKey"]))

        out = io.BytesIO()
        pr.write(out, only_new=True)
        self.assertEqual(out.getvalue(), b"NewKey: new new\n")

    def test_import_dict(self):
        """Import a dictionary with update()."""
        pr = problem_report.ProblemReport()
        pr["oldtext"] = "Hello world"
        pr["oldbin"] = bin_data
        pr["overwrite"] = "I am crap"

        d = {
            "newtext": "Goodbye world",
            "newbin": "11\000\001\002\xFFZZ",
            "overwrite": "I am good",
        }

        pr.update(d)
        self.assertEqual(pr["oldtext"], "Hello world")
        self.assertEqual(pr["oldbin"], bin_data)
        self.assertEqual(pr["newtext"], "Goodbye world")
        self.assertEqual(pr["newbin"], "11\000\001\002\xFFZZ")
        self.assertEqual(pr["overwrite"], "I am good")

    def test_load_key_filter(self):
        """Load a report with filtering keys."""
        report = textwrap.dedent(
            """\
            ProblemType: Crash
            DataNo: nonono
            GoodFile: base64
             H4sICAAAAAAC/0FmaWxlAA==
             c3RyhEIGBoYoRiYAM5XUCxAAAAA=
            DataYes: yesyes
            BadFile: base64
             H4sICAAAAAAC/0ZpbGUA
             S8vPZ0hKLAIACq50HgcAAAA=
            """
        ).encode()
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report), key_filter=["DataYes", "GoodFile"])
        self.assertEqual(pr["DataYes"], "yesyes")
        self.assertEqual(pr["GoodFile"], bin_data)
        self.assertEqual(sorted(pr.keys()), ["DataYes", "GoodFile"])

    def test_get_on_disk_size(self) -> None:
        """Test CompressedValue.get_on_disk_size()."""
        compressed_value = problem_report.CompressedValue(b"somedata")
        self.assertEqual(compressed_value.get_compressed_size(), 28)
        assert compressed_value.compressed_value is not None
        base64_encoded = base64.b64encode(compressed_value.compressed_value)
        self.assertEqual(compressed_value.get_on_disk_size(), len(base64_encoded))


class TestEntryParser(unittest.TestCase):
    # pylint: disable=protected-access
    """Test _EntryParser class."""

    def test_parse(self) -> None:
        """Test parsing a report file with _EntryParser."""
        report_file = io.BytesIO(
            textwrap.dedent(
                """\
                First: single line
                Second: multi
                 line
                """
            ).encode()
        )
        entries = []
        for entry in problem_report._EntryParser(report_file):
            entries.append(list(entry))
        self.assertEqual(
            entries, [[b"First: single line\n"], [b"Second: multi\n", b" line\n"]]
        )

    def test_skip_entries(self) -> None:
        """Test skipping reading one entry."""
        report_file = io.BytesIO(
            textwrap.dedent(
                """\
                First: this entry
                 will be skipped
                Second: single line
                """
            ).encode()
        )
        iterator = problem_report._EntryParser(report_file)
        next(iterator)
        self.assertEqual(list(next(iterator)), [b"Second: single line\n"])

    def test_skip_partial_entries(self) -> None:
        """Test skipping reading one entry."""
        report_file = io.BytesIO(
            textwrap.dedent(
                """\
                First: this line is read,
                 but these lines
                 will be skipped
                Second: single line
                """
            ).encode()
        )
        iterator = problem_report._EntryParser(report_file)
        self.assertEqual(next(next(iterator)), b"First: this line is read,\n")
        self.assertEqual(list(next(iterator)), [b"Second: single line\n"])

    def test_skip_last_entry(self) -> None:
        """Test skipping reading the last entry."""
        report_file = io.BytesIO(
            textwrap.dedent(
                """\
                First: this line is read
                Second: this line is skipped
                """
            ).encode()
        )
        iterator = problem_report._EntryParser(report_file)
        self.assertEqual(list(next(iterator)), [b"First: this line is read\n"])
        next(iterator)
        with self.assertRaises(StopIteration):
            next(iterator)



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Integration tests for whoopsie-upload-all."""

import io
import os
import shutil
import tempfile
import unittest
import unittest.mock

from tests.helper import import_module_from_file
from tests.paths import get_data_directory

whoopsie_upload_all = import_module_from_file(
    get_data_directory() / "whoopsie-upload-all"
)


class TestWhoopsieUploadAll(unittest.TestCase):
    """Integration tests for whoopsie-upload-all."""

    def setUp(self):
        self.report_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.report_dir)

    def _write_report(self, content: bytes) -> str:
        report = os.path.join(self.report_dir, "testcase.crash")
        with open(report, "wb") as report_file:
            report_file.write(content)
        return report

    @unittest.mock.patch("sys.stderr", new_callable=io.StringIO)
    def test_process_report_malformed_report(self, stderr_mock):
        """Test process_report() raises MalformedProblemReport."""
        report = self._write_report(b"AB\xfc:CD\n")
        self.assertIsNone(whoopsie_upload_all.process_report(report))
        self.assertIn(
            "Malformed problem report: 'ascii' codec can't decode byte 0xfc"
            " in position 2: ordinal not in range(128).",
            stderr_mock.getvalue(),
        )



# Copyright (C) 2023 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.unicode_gettext."""

import unittest
import unittest.mock

from apport import unicode_gettext


class TestDeprecation(unittest.TestCase):
    """Unit tests for apport.unicode_gettext."""

    def test_unicode_gettext(self) -> None:
        """unicode_gettext() throws a deprecation warning."""
        with self.assertWarns(PendingDeprecationWarning):
            self.assertEqual(unicode_gettext("untranslated"), "untranslated")



"""Helper functions for the test cases."""

import contextlib
import importlib.machinery
import importlib.util
import os
import pathlib
import shutil
import subprocess
import time
import unittest.mock
from collections.abc import Callable, Generator, Iterator, Sequence, Set
from typing import Any
from unittest.mock import MagicMock

import psutil


def get_init_system() -> str:
    """Return the name of the running init system (PID 1)."""
    with open("/proc/1/comm", encoding="utf-8") as comm:
        return comm.read().rstrip()


def import_module_from_file(path: pathlib.Path) -> Any:
    """Import a module by its filename."""
    name = path.stem.replace("-", "_")
    spec = importlib.util.spec_from_loader(
        name, importlib.machinery.SourceFileLoader(name, str(path))
    )
    assert spec is not None and spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@contextlib.contextmanager
def pidfd_open(pid: int) -> Iterator[int]:
    """Return a file descriptor referring to the process pid.

    This function provides os.pidfd_open() as context manager."""
    pidfd = os.pidfd_open(pid)
    try:
        yield pidfd
    finally:
        os.close(pidfd)


def pidof(program):
    """Find the process ID of a running program.

    This function wraps the pidof command and returns a set of
    process IDs.
    """
    try:
        stdout = subprocess.check_output(["/bin/pidof", "-nx", program])
    except subprocess.CalledProcessError as error:
        if error.returncode == 1:
            return set()
        raise  # pragma: no cover
    return set(int(pid) for pid in stdout.decode().split())


def read_shebang(command: str) -> str | None:
    """Return the shebang of the given file.

    If the given file is a script, return the executable from the
    shebang. Otherwise `None`.
    """
    with open(command, "rb") as command_file:
        first_line = command_file.readline(100).strip()
    if not first_line.startswith(b"#!"):
        return None
    return first_line.decode().split(" ", 1)[0][2:]


@contextlib.contextmanager
def restore_os_environ() -> Generator[None, None, None]:
    """Restore os.environ after leaving this context manager."""
    orig_env = os.environ.copy()
    yield
    os.environ.clear()
    os.environ.update(orig_env)


@contextlib.contextmanager
def run_test_executable(
    args: Sequence[str] = (os.path.realpath("/bin/sleep"), "86400"),
    env: dict[str, str] | None = None,
) -> Iterator[int]:
    """Run test executable and yield the process ID. Kill process afterwards."""
    with subprocess.Popen(args, env=env) as test_process:
        try:
            yield test_process.pid
        finally:
            test_process.kill()


def _id(obj):
    return obj


def skip_if_command_is_missing(cmd: str) -> Callable:
    """Skip a test if the command is not found."""
    if shutil.which(cmd) is None:
        return unittest.skip(f"{cmd} not installed")
    return _id


@contextlib.contextmanager
def wrap_object(
    target: object, attribute: str, include_instance: bool = False
) -> Generator[MagicMock, None, None]:
    """Wrap the named member on an object with a mock object.

    wrap_object() can be used as a context manager. Inside the
    body of the with statement, the attribute of the target is
    wrapped with a :class:`unittest.mock.MagicMock` object. When
    the with statement exits the patch is undone.

    The instance argument 'self' of the wrapped attribute will
    not be logged in the MagicMock call if include_instance is
    set to False. This allows using the assert calls on the mock
    without differentiating between different instances.

    See also https://stackoverflow.com/questions/44768483 for
    the use case.
    """
    mock = MagicMock()
    real_attribute = getattr(target, attribute)

    def mocked_attribute(self, *args, **kwargs):
        if include_instance:
            mock(self, *args, **kwargs)
        else:
            mock(*args, **kwargs)
        return real_attribute(self, *args, **kwargs)

    with unittest.mock.patch.object(target, attribute, mocked_attribute):
        yield mock


def wait_for_sleeping_state(pid: int, timeout: float = 5.0) -> None:
    """Wait for sleep command to enter sleeping state."""
    proc = psutil.Process(pid)
    waited = 0.0
    last_state = ""
    while waited < timeout:
        last_state = proc.status()
        if last_state == "sleeping":
            return

        time.sleep(0.1)
        waited += 0.1

    raise TimeoutError(
        f"{pid=} did not enter 'sleeping' state after {timeout=} seconds."
        f" Got {last_state!r} instead."
    )


def wait_for_process_to_appear(
    process: str, already_running: Set[int], timeout: float = 5.0
) -> int:
    """Wait for process to appear and return its PID."""
    waited = 0.0
    while waited < timeout:
        pids = pidof(process) - already_running
        if pids:
            assert len(pids) == 1, f"Found more than one PID for {process!r}"
            return pids.pop()
        time.sleep(0.1)
        timeout -= 0.1

    raise TimeoutError(f"PID for {process!r} not found within {timeout=} seconds.")



"""Abstract Apport user interface.

This encapsulates the workflow and common code for any user interface
implementation (like GTK, Qt, or CLI).
"""

# Copyright (C) 2007 - 2011 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import argparse
import ast
import configparser
import dataclasses
import errno
import gettext
import glob
import gzip
import io
import locale
import os.path
import pwd
import queue
import re
import shutil
import signal
import smtplib
import subprocess
import sys
import threading
import time
import traceback
import urllib.error
import webbrowser
import zlib
from collections.abc import Sequence
from gettext import gettext as _
from typing import Any

import apport.crashdb
import apport.fileutils
import apport.logging
import apport.REThread
from apport.packaging_impl import impl as packaging
from apport.user_group import get_process_user_and_group

__version__ = "2.28.3"


symptom_script_dir = os.environ.get("APPORT_SYMPTOMS_DIR", "/usr/share/apport/symptoms")
PF_KTHREAD = 0x200000


def get_pid(report):
    """Extract process ID from report."""
    try:
        pid = re.search("Pid:\t(.*)\n", report.get("ProcStatus", "")).group(1)
        return int(pid)
    except (IndexError, AttributeError):
        return None


def _get_env_int(key: str, default: int | None = None) -> int | None:
    """Get an environment variable as integer.

    Return None if it doesn't exist or failed to convert to integer.
    The optional second argument can specify an alternate default.
    """
    try:
        return int(os.environ[key])
    except (KeyError, ValueError):
        return default


def _get_newest_process_for_user(name: str, uid: int) -> int | None:
    process = subprocess.run(
        ["pgrep", "-n", "-x", "-u", str(uid), name],
        capture_output=True,
        check=False,
        text=True,
    )
    if process.returncode != 0 or not process.stdout:
        return None
    return int(process.stdout.strip())


def _get_users_environ(uid: int) -> dict[str, str]:
    """Find D-BUS address and XDG_DATA_DIRS for the given user.

    The D-BUS address and XDG_DATA_DIRS is needed for xdg-open. It is
    incredibly hard, or alternatively, unsafe to funnel it through
    pkexec/env/sudo, so grab it from gvfsd.
    """
    gvfsd_pid = _get_newest_process_for_user("gvfsd", uid)
    if gvfsd_pid is None:
        return {}

    gvfsd_pid_fd = os.open(
        f"/proc/{gvfsd_pid}", os.O_RDONLY | os.O_PATH | os.O_DIRECTORY
    )
    try:
        gvfsd_env = apport.fileutils.get_process_environ(gvfsd_pid_fd)
    except OSError:
        return {}
    finally:
        os.close(gvfsd_pid_fd)

    return {
        key: gvfsd_env[key]
        for key in ("DBUS_SESSION_BUS_ADDRESS", "XDG_DATA_DIRS")
        if key in gvfsd_env
    }


def has_display() -> bool:
    """Check if a display server is specified in the environment."""
    return bool(os.environ.get("DISPLAY") or os.environ.get("WAYLAND_DISPLAY"))


def run_as_real_user(
    args: list[str], *, get_user_env: bool = False, **kwargs: Any
) -> None:
    """Call subprocess.run as real user if called via sudo/pkexec.

    If we are called through pkexec/sudo, determine the real user ID and
    run the command with it to get the user's web browser settings.
    If get_user_env is set to True, the D-BUS address and XDG_DATA_DIRS
    is grabbed from a running gvfsd and added to the process environment.
    """
    uid = _get_env_int("SUDO_UID", _get_env_int("PKEXEC_UID"))
    if uid is None or not get_process_user_and_group().is_root():
        subprocess.run(args, check=False, **kwargs)
        return

    pwuid = pwd.getpwuid(uid)

    gid = _get_env_int("SUDO_GID")
    if gid is None:
        gid = pwuid.pw_gid

    env = {
        k: v
        for k, v in os.environ.items()
        if not k.startswith("SUDO_") and k != "PKEXEC_UID"
    }
    if get_user_env:
        env |= _get_users_environ(uid)
    env["HOME"] = pwuid.pw_dir
    subprocess.run(
        args,
        check=False,
        env=env,
        user=uid,
        group=gid,
        extra_groups=os.getgrouplist(pwuid.pw_name, gid),
        **kwargs,
    )


def still_running(pid):
    """Check if the process with the given ID is still running."""
    try:
        os.kill(int(pid), 0)
    except OSError as error:
        if error.errno == errno.ESRCH:
            return False
    return True


def thread_collect_info(
    report, reportfile, package, ui, symptom_script=None, ignore_uninstalled=False
):
    # TODO: Split into smaller functions/methods
    # pylint: disable=too-many-branches,too-many-statements
    """Collect information about report.

    Encapsulate calls to add_*_info() and update given report, so that this
    function is suitable for threading.

    ui must be a HookUI instance, it gets passed to add_hooks_info().

    If reportfile is not None, the file is written back with the new data.

    If symptom_script is given, it will be run first (for run_symptom()).
    """
    assert isinstance(ui, HookUI)
    try:
        report.add_gdb_info()
    except OSError:
        # it's okay if gdb is not installed on the client side; we'll get stack
        # traces on retracing.
        pass
    report.add_os_info()

    if symptom_script:
        symb = {}
        try:
            with open(symptom_script, encoding="utf-8") as f:
                # legacy, pylint: disable=exec-used
                exec(compile(f.read(), symptom_script, "exec"), symb)
            package = symb["run"](report, ui)
            if not package:
                apport.logging.error(
                    "symptom script %s did not determine the affected package",
                    symptom_script,
                )
                return
            report["Symptom"] = os.path.splitext(os.path.basename(symptom_script))[0]
        except StopIteration:
            sys.exit(0)
        except Exception:  # pylint: disable=broad-except
            apport.logging.error("symptom script %s crashed:", symptom_script)
            traceback.print_exc()
            sys.exit(0)

    if not package:
        if "ExecutablePath" in report:
            package = apport.fileutils.find_file_package(report["ExecutablePath"])
        else:
            raise KeyError(
                "called without a package, and report does not have ExecutablePath"
            )

    # check if the package name relates to an installed snap
    snap = apport.fileutils.find_snap(package)
    if snap:
        report.add_snap_info(snap)

    try:
        report.add_package_info(package)
    except ValueError:
        # this happens if we are collecting information on an uninstalled
        # package

        # we found no package, but a snap, so lets continue
        if not ignore_uninstalled and "Snap" not in report:
            raise
    except SystemError as error:
        report["UnreportableReason"] = str(error)

    if "UnreportableReason" not in report:
        if report.add_hooks_info(ui):
            sys.exit(0)

        # check package origin; we do that after adding hooks, so that hooks
        # have the chance to set a third-party CrashDB.
        try:
            if (
                "CrashDB" not in report
                and "APPORT_DISABLE_DISTRO_CHECK" not in os.environ
            ):
                if "Package" not in report and "Snap" not in report:
                    report["UnreportableReason"] = _(
                        "This package does not seem to be installed correctly"
                    )
                elif not packaging.is_distro_package(
                    report["Package"].split()[0]
                ) and not packaging.is_native_origin_package(
                    report["Package"].split()[0]
                ):
                    # TRANS: %s is the name of the operating system
                    report["UnreportableReason"] = (
                        _(
                            "This does not seem to be an official %s package."
                            " Please retry after updating the indexes of"
                            " available packages, if that does not work"
                            " then remove related third party packages"
                            " and try again."
                        )
                        % report["DistroRelease"].split()[0]
                    )
        except ValueError:
            # this happens if we are collecting information on an uninstalled
            # package

            # we found no package, but a snap, so lets continue
            if not ignore_uninstalled and "Snap" not in report:
                raise

    # add title
    if "Title" not in report:
        title = report.standard_title()
        if title:
            report["Title"] = title

    # check obsolete packages
    if (
        report.get("ProblemType") == "Crash"
        and "APPORT_IGNORE_OBSOLETE_PACKAGES" not in os.environ
    ):
        old_pkgs = report.obsolete_packages()
        if old_pkgs:
            report["UnreportableReason"] = _(
                "You have some obsolete package versions installed."
                " Please upgrade the following packages"
                " and check if the problem still occurs:\n\n%s"
            ) % ", ".join(old_pkgs)

    if reportfile:
        try:
            with open(reportfile, "ab") as f:
                os.chmod(reportfile, 0)
                report.write(f, only_new=True)
        except OSError as error:
            # this should happen very rarely; presumably a new crash report is
            # being generated by a background apport instance (which will set
            # the file to permissions zero while writing), while the first
            # report is being processed
            apport.logging.error("Cannot update %s: %s", reportfile, error)

        apport.fileutils.mark_report_seen(reportfile)
        os.chmod(reportfile, 0o640)


@dataclasses.dataclass
class Action:
    """Action to take on a problem report.

    Possible actions: examine the crash ('examine'), report the crash
    ('report'), restart the crashed application ('restart'), or ignore further
    crashes ('ignore').
    """

    examine: bool = False
    ignore: bool = False
    remember: bool = False
    report: bool = False
    restart: bool = False


class UserInterface:
    # TODO: Check if some methods can be made private
    # pylint: disable=too-many-public-methods
    """Apport user interface API.

    This provides an abstract base class for encapsulating the workflow and
    common code for any user interface implementation (like GTK, Qt, or CLI).

    A concrete subclass must implement all the abstract ui_* methods.
    """

    def __init__(self, argv: list[str]):
        """Initialize program state and parse command line options."""
        self.gettext_domain = "apport"
        self.report: apport.report.Report | None = None
        self.report_file: str | None = None
        self.cur_package = None
        self.offer_restart = False
        self.specified_a_pkg = False
        self.upload_progress = None

        try:
            self.crashdb = apport.crashdb.get_crashdb(None)
        except ImportError as error:
            # this can happen while upgrading python packages
            apport.logging.fatal(
                "Could not import module, is a package upgrade in progress?"
                " Error: %s",
                str(error),
            )
        except KeyError:
            apport.logging.fatal(
                "/etc/apport/crashdb.conf is damaged: No default database"
            )

        gettext.textdomain(self.gettext_domain)
        self.args = self.parse_argv(argv)

    #
    # main entry points
    #

    def run_crashes(self):
        """Present all currently pending crash reports.

        Ask the user what to do about them, and offer to file bugs for them.

        Crashes that occurred in a different desktop (logind) session than the
        one that is currently running are not processed. This skips crashes
        that happened during logout, which are uninteresting and confusing to
        see at the next login.

        Return True if at least one crash report was processed, False
        otherwise.
        """
        result = False
        # for iterating over /var/crash (as opposed to running on or clicking
        # on a particular .crash file) we offer restarting
        self.offer_restart = True

        if os.geteuid() == 0:
            reports = apport.fileutils.get_new_system_reports()
        else:
            reports = apport.fileutils.get_new_reports()

        for f in reports:
            if not self.load_report(f):
                continue
            assert self.report

            if self.report["ProblemType"] == "Hang":
                self.finish_hang(f)
            else:
                self.run_crash(f)
            result = True

        return result

    def run_crash(self, report_file):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-return-statements
        # pylint: disable=too-many-statements
        """Present and report a particular crash.

        If confirm is True, ask the user what to do about it, and offer to file
        a bug for it.

        If confirm is False, the user will not be asked, and the crash is
        reported right away.
        """
        self.report_file = report_file

        try:
            try:
                apport.fileutils.mark_report_seen(report_file)
            except OSError:
                # not there any more? no problem, then it won't be regarded as
                # "seen" any more anyway
                pass
            if not self.report and not self.load_report(report_file):
                return
            assert self.report

            if "Ignore" in self.report:
                return

            # check for absent CoreDumps (removed if they exceed size limit)
            if (
                self.report.get("ProblemType") == "Crash"
                and "Signal" in self.report
                and "CoreDump" not in self.report
                and "Stacktrace" not in self.report
            ):
                subject = os.path.basename(
                    self.report.get("ExecutablePath", _("unknown program"))
                )
                heading = _('Sorry, the program "%s" closed unexpectedly') % subject
                footer = _(
                    "Your computer does not have enough free "
                    "memory to automatically analyze the problem "
                    "and send a report to the developers."
                )
                self.ui_error_message(
                    _("Problem in %s") % subject, f"{heading}\n\n{footer}"
                )
                return

            allowed_to_report = apport.fileutils.allowed_to_report()
            response = self.ui_present_report_details(allowed_to_report)
            if response.report or response.examine:
                if "_MarkForUpload" not in self.report:
                    self.collect_info()

            if self.report is None:
                # collect() does that on invalid reports
                return

            if response.examine:
                self.examine()
                return
            if response.restart:
                self.restart()
            if response.ignore:
                self.report.mark_ignore()
            try:
                if response.remember:
                    self.remember_send_report(response.report)
            # use try/expect for python2 support. Old reports (generated
            # pre-apport 2.20.10-0ubuntu4) may not have the remember key
            # and can be loaded afterwards (or after dist-upgrade)
            except KeyError:
                pass
            if not response.report:
                return

            # We don't want to send crashes to the crash database for binaries
            # that changed since the crash happened. See LP: #1039220 for
            # details.
            if (
                "_MarkForUpload" in self.report
                and self.report["_MarkForUpload"] != "False"
            ):
                apport.fileutils.mark_report_upload(report_file)
            # We check for duplicates and unreportable crashes here, rather
            # than before we show the dialog, as we want to submit these to the
            # crash database, but not Launchpad.
            if self.crashdb.accepts(self.report):
                # FIXME: This behaviour is not really correct, but necessary as
                # long as we only support a single crashdb and have whoopsie
                # hardcoded. Once we have multiple crash dbs, we need to check
                # accepts() earlier, and not even present the data if none of
                # the DBs wants the report. See LP#957177 for details.
                if self.handle_duplicate():
                    return
                if self.check_unreportable():
                    return
                self.file_report()
        except PermissionError:
            self.ui_error_message(
                _("Invalid problem report"),
                _("You are not allowed to access this problem report."),
            )
            sys.exit(1)
        except OSError as error:
            if error.errno == errno.ENOMEM:
                apport.logging.fatal("Out of memory, aborting")
            elif error.errno == errno.ENOSPC:
                self.ui_error_message(
                    _("Error"),
                    _(
                        "There is not enough disk space"
                        " available to process this report."
                    ),
                )
                sys.exit(1)
            elif error.errno == errno.EIO:
                self.ui_error_message(_("Invalid problem report"), error.strerror)
                sys.exit(1)
            raise

    @staticmethod
    def finish_hang(f):
        """Finish processing a hanging application after the core pipe handler
        has handed the report back.

        This will signal to whoopsie that the report needs to be uploaded.
        """
        apport.fileutils.mark_report_upload(f)
        apport.fileutils.mark_report_seen(f)

    def run_hang(self, pid):
        """Report an application hanging.

        This will first present a dialog containing the information it can
        collect from the running application (everything but the trace) with
        the option of terminating or restarting the application, optionally
        reporting that this error occurred.

        A SIGABRT will then be sent to the process and a series of
        noninteractive processes will collect the remaining information and
        mark the report for uploading.
        """
        self.report = apport.Report("Hang")

        if not self.args.pid:
            self.ui_error_message(
                _("No PID specified"),
                _("You need to specify a PID. See --help for more information."),
            )
            return False

        try:
            self.report.add_proc_info(pid)
        except ValueError as error:
            if str(error) == "invalid process":
                self.ui_error_message(
                    _("Invalid PID"), _("The specified process ID does not exist.")
                )
                sys.exit(1)
            elif str(error) == "not accessible":
                self.ui_error_message(
                    _("Not your PID"),
                    _("The specified process ID does not belong to you."),
                )
                sys.exit(1)
        self.report.add_package_info()
        path = self.report.get("ExecutablePath", "")
        self.cur_package = apport.fileutils.find_file_package(path)
        self.report.add_os_info()
        allowed_to_report = apport.fileutils.allowed_to_report()
        response = self.ui_present_report_details(allowed_to_report, modal_for=int(pid))
        if response.report:
            apport.fileutils.mark_hanging_process(self.report, pid)
            os.kill(int(pid), signal.SIGABRT)
        else:
            os.kill(int(pid), signal.SIGKILL)

        if response.restart:
            self.wait_for_pid(pid)
            self.restart()
        return True

    @staticmethod
    def wait_for_pid(pid):
        """waitpid() does not work for non-child processes. Query the process
        state in a loop, waiting for "no such process."
        """
        while True:
            try:
                os.kill(int(pid), 0)
            except OSError as error:
                if error.errno == errno.ESRCH:
                    break
                raise
            time.sleep(1)

    @staticmethod
    def kill_segv(pid):
        """Kill process with signal SIGSEGV."""
        os.kill(int(pid), signal.SIGSEGV)

    def run_report_bug(self, symptom_script: str | None = None) -> bool:
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-return-statements
        # pylint: disable=too-many-statements
        """Report a bug.

        If a pid is given on the command line, the report will contain runtime
        debug information. Either a package or a pid must be specified; if none
        is given, show a list of symptoms.

        If a symptom script is given, this will be run first (used by
        run_symptom()).
        """
        if not self.args.package and not self.args.pid and not symptom_script:
            if self.run_symptoms():
                return True

            self.ui_error_message(
                _("No package specified"),
                _(
                    "You need to specify a package or a PID."
                    " See --help for more information."
                ),
            )
            return False

        self.report = apport.Report("Bug")

        # if PID is given, add info
        if self.args.pid:
            try:
                proc_pid_fd = os.open(
                    f"/proc/{self.args.pid}", os.O_RDONLY | os.O_PATH | os.O_DIRECTORY
                )
                stat_file = os.open("stat", os.O_RDONLY, dir_fd=proc_pid_fd)
                with io.open(stat_file, encoding="utf-8") as f:
                    stat = f.read().split()
                flags = int(stat[8])
                if flags & PF_KTHREAD:
                    # this PID is a kernel thread
                    self.args.package = "linux"
                else:
                    self.report.add_proc_info(
                        pid=self.args.pid, proc_pid_fd=proc_pid_fd
                    )
            except PermissionError:
                self.ui_error_message(
                    _("Permission denied"),
                    _(
                        "The specified process does not belong to you. Please"
                        " run this program as the process owner or as root."
                    ),
                )
                return False
            except (ValueError, OSError) as error:
                if getattr(error, "errno", None) == errno.ENOENT:
                    # silently ignore nonexisting PIDs; the user must not
                    # close the application prematurely
                    return False
                self.ui_error_message(
                    _("Invalid PID"),
                    _("The specified process ID does not belong to a program."),
                )
                return False
        else:
            self.report.add_proc_environ()

        if self.args.package:
            self.args.package = self.args.package.strip()
        # "Do what I mean" for filing against "linux"
        if self.args.package == "linux":
            self.cur_package = packaging.get_kernel_package()
        else:
            self.cur_package = self.args.package

        try:
            self.collect_info(symptom_script)
        except ValueError as error:
            if "package" in str(error) and "does not exist" in str(error):
                if not self.cur_package:
                    self.ui_error_message(
                        _("Invalid problem report"),
                        _("Symptom script %s did not determine an affected package")
                        % symptom_script,
                    )
                else:
                    self.ui_error_message(
                        _("Invalid problem report"),
                        _("Package %s does not exist") % self.cur_package,
                    )
                return False
            raise

        if self.check_unreportable():
            return False

        self.add_extra_tags()

        if self.handle_duplicate():
            return True

        # not useful for bug reports, and has potentially sensitive information
        try:
            del self.report["ProcCmdline"]
        except KeyError:
            pass

        if self.args.save:
            try:
                savefile = os.path.expanduser(self.args.save)
                if savefile.endswith(".gz"):
                    with gzip.open(savefile, "wb") as f:
                        self.report.write(f)
                else:
                    with open(os.path.expanduser(self.args.save), "wb") as f:
                        self.report.write(f)
            except OSError as error:
                self.ui_error_message(_("Cannot create report"), str(error))
        else:
            # show what's being sent
            allowed_to_report = True
            response = self.ui_present_report_details(allowed_to_report)
            if response.report:
                self.file_report()

        return True

    def run_update_report(self) -> bool:
        """Update an existing bug with locally collected information."""
        # avoid irrelevant noise
        if not self.crashdb.can_update(self.args.update_report):
            self.ui_error_message(
                _("Updating problem report"),
                _(
                    "You are not the reporter or subscriber of this "
                    "problem report, or the report is a duplicate or already "
                    'closed.\n\nPlease create a new report using "apport-bug".'
                ),
            )
            return False

        is_reporter = self.crashdb.is_reporter(self.args.update_report)

        if not is_reporter:
            r = self.ui_question_yesno(
                _(
                    "You are not the reporter of this problem report. It is "
                    "much easier to mark a bug as a duplicate of another than "
                    "to move your comments and attachments to a new bug.\n\n"
                    "Subsequently, we recommend that you file a new bug "
                    'report using "apport-bug" and make a comment in this bug '
                    "about the one you file.\n\nDo you really want to proceed?"
                )
            )
            if not r:
                return False

        # list of affected source packages
        self.report = apport.Report("Bug")
        if self.args.package:
            pkgs = [self.args.package.strip()]
        else:
            pkgs = self.crashdb.get_affected_packages(self.args.update_report)

        info_collected = False
        for p in pkgs:
            # print(f"Collecting apport information for source package {p}...")
            self.cur_package = p
            self.report["SourcePackage"] = p
            self.report["Package"] = p  # no way to find this out

            # we either must have the package installed or a source package
            # hook available to collect sensible information
            try:
                packaging.get_version(p)
            except ValueError:
                if not os.path.exists(
                    os.path.join(apport.report.PACKAGE_HOOK_DIR, f"source_{p}.py")
                ):
                    print(f"Package {p} not installed and no hook available, ignoring")
                    continue
            self.collect_info(ignore_uninstalled=True)
            info_collected = True

        if not info_collected:
            self.ui_info_message(
                _("Updating problem report"), _("No additional information collected.")
            )
            return False

        self.report.add_user_info()
        self.report.add_proc_environ()
        self.add_extra_tags()

        # delete the uninteresting keys
        del self.report["Date"]
        try:
            del self.report["SourcePackage"]
        except KeyError:
            pass

        if len(self.report) == 0:
            self.ui_info_message(
                _("Updating problem report"), _("No additional information collected.")
            )
            return False

        # show what's being sent
        allowed_to_report = True
        response = self.ui_present_report_details(allowed_to_report)
        if response.report:
            self.crashdb.update(
                self.args.update_report,
                self.report,
                "apport information",
                change_description=is_reporter,
                attachment_comment="apport information",
            )
            return True

        return False

    def run_symptoms(self):
        """Report a bug from a list of available symptoms.

        Return False if no symptoms are available.
        """
        scripts = glob.glob(os.path.join(symptom_script_dir, "*.py"))

        # symptoms contains a list of (symptom_description, symptom_name)
        symptoms = []
        for script in scripts:
            # scripts with an underscore can be used for private libraries
            if os.path.basename(script).startswith("_"):
                continue
            symb = {}
            try:
                with open(script, encoding="utf-8") as f:
                    # legacy, pylint: disable=exec-used
                    exec(compile(f.read(), script, "exec"), symb)
            except Exception:  # pylint: disable=broad-except
                apport.logging.error("symptom script %s is invalid", script)
                traceback.print_exc()
                continue
            if "run" not in symb:
                apport.logging.error(
                    "symptom script %s does not define run() function", script
                )
                continue
            symptom_name = os.path.splitext(os.path.basename(script))[0]
            symptoms.append((symb.get("description", symptom_name), symptom_name))

        if not symptoms:
            return False

        symptoms.sort()
        symptoms.append(("Other problem", None))

        ch = self.ui_question_choice(
            _("What kind of problem do you want to report?"),
            [description for description, name in symptoms],
            False,
        )

        if ch is not None:
            symptom = symptoms[ch[0]][1]
            if symptom:
                self.run_report_bug(os.path.join(symptom_script_dir, symptom + ".py"))
            else:
                return False

        return True

    def run_symptom(self):
        """Report a bug with a symptom script."""
        script = os.path.join(symptom_script_dir, self.args.symptom + ".py")
        if not os.path.exists(script):
            self.ui_error_message(
                _("Unknown symptom"),
                _('The symptom "%s" is not known.') % self.args.symptom,
            )
            return

        self.run_report_bug(script)

    def run_argv(self):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-return-statements
        """Call appropriate run_* method according to command line arguments.

        Return True if at least one report has been processed, and False
        otherwise.
        """
        if self.args.symptom:
            self.run_symptom()
            return True
        if self.args.hanging:
            self.run_hang(self.args.pid)
            return True
        if self.args.filebug:
            return self.run_report_bug()
        if self.args.update_report is not None:
            return self.run_update_report()
        if self.args.version:
            print(__version__)
            return True
        if self.args.crash_file:
            try:
                self.run_crash(self.args.crash_file)
            except OSError as error:
                self.ui_error_message(_("Invalid problem report"), str(error))
            return True
        if self.args.window:
            if os.getenv("XDG_SESSION_TYPE") == "wayland":
                self.ui_error_message(
                    _("Cannot create report"),
                    _(
                        "The window option cannot be used on "
                        "Wayland.\n\nPlease find the window's "
                        "process ID and then run 'ubuntu-bug "
                        "<process ID>'."
                        "\n\nThe process ID can be found "
                        "by running the System Monitor application. "
                        "In the Processes tab, scroll until you "
                        "find the correct application. The process "
                        "ID is the number listed in the ID column."
                    ),
                )
                return True

            self.ui_info_message(
                "",
                _(
                    "After closing this message please click on an"
                    " application window to report a problem about it."
                ),
            )
            xprop = subprocess.run(
                ["xprop", "_NET_WM_PID"],
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if xprop.returncode == 0:
                try:
                    self.args.pid = int(xprop.stdout.split()[-1])
                except ValueError:
                    self.ui_error_message(
                        _("Cannot create report"),
                        _("xprop failed to determine process ID of the window"),
                    )
                    return True
                return self.run_report_bug()

            self.ui_error_message(
                _("Cannot create report"),
                _("xprop failed to determine process ID of the window")
                + "\n\n"
                + xprop.stderr.decode(),
            )
            return True

        return self.run_crashes()

    #
    # methods that implement workflow bits
    #

    @staticmethod
    def parse_argv_update(argv: Sequence[str]) -> argparse.Namespace:
        """Parse command line options when being invoked in update mode."""
        parser = argparse.ArgumentParser(usage=_("%(prog)s <report number>"))
        parser.add_argument("-p", "--package", help=_("Specify package name."))
        parser.add_argument(
            "--tag",
            action="append",
            default=[],
            dest="tags",
            help=_("Add an extra tag to the report. Can be specified multiple times."),
        )
        parser.add_argument("update_report", metavar="report_number", type=int)
        args = parser.parse_args(argv[1:])

        args.symptom = None
        args.filebug = False
        args.crash_file = None
        args.version = False
        args.hanging = False
        return args

    def parse_argv(self, argv: list[str]) -> argparse.Namespace:
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-statements
        """Parse command line options.

        If a single argument is given without any options, this tries to "do
        what I mean".
        """
        # invoked in update mode?
        if len(argv) > 0:
            if "APPORT_INVOKED_AS" in os.environ:
                argv[0] = os.path.join(
                    os.path.dirname(argv[0]),
                    os.path.basename(os.environ["APPORT_INVOKED_AS"]),
                )
            cmd = argv[0]
            if cmd.endswith("-update-bug") or cmd.endswith("-collect"):
                return self.parse_argv_update(argv)

        if len(argv) > 0 and cmd.endswith("-bug"):
            suppress = argparse.SUPPRESS
        else:
            suppress = None

        parser = argparse.ArgumentParser(
            usage=_(
                "%(prog)s [options]"
                " [symptom|pid|package|program path|.apport/.crash file]"
            )
        )
        parser.add_argument(
            "-f",
            "--file-bug",
            action="store_true",
            dest="filebug",
            help=suppress
            or _(
                "Start in bug filing mode. Requires --package and an optional"
                " --pid, or just a --pid. If neither is given, display a list"
                " of known symptoms. (Implied if a single argument is given.)"
            ),
        )
        parser.add_argument(
            "-w",
            "--window",
            action="store_true",
            help=_("Click a window as a target for filing a problem report."),
        )
        parser.add_argument(
            "-u",
            "--update-bug",
            type=int,
            dest="update_report",
            help=suppress
            or _("Start in bug updating mode. Can take an optional --package."),
        )
        parser.add_argument(
            "-s",
            "--symptom",
            metavar="SYMPTOM",
            help=suppress
            or _(
                "File a bug report about a symptom. (Implied if symptom name"
                " is given as only argument.)"
            ),
        )
        parser.add_argument(
            "-p",
            "--package",
            help=suppress
            or _(
                "Specify package name in --file-bug mode. This is optional"
                " if a --pid is specified. (Implied if package name"
                " is given as only argument.)"
            ),
        )
        parser.add_argument(
            "-P",
            "--pid",
            type=int,
            help=suppress
            or _(
                "Specify a running program in --file-bug mode. If this is"
                " specified, the bug report will contain more information."
                "  (Implied if pid is given as only argument.)"
            ),
        )
        parser.add_argument(
            "--hanging",
            action="store_true",
            help=_("The provided pid is a hanging application."),
        )
        parser.add_argument(
            "-c",
            "--crash-file",
            metavar="PATH",
            help=suppress
            or _(
                "Report the crash from given .apport or .crash file"
                " instead of the pending ones in %s."
                " (Implied if file is given as only argument.)"
            )
            % apport.fileutils.report_dir,
        )
        parser.add_argument(
            "--save",
            metavar="PATH",
            help=_(
                "In bug filing mode, save the collected information into"
                " a file instead of reporting it. This file can then be"
                " reported later on from a different machine."
            ),
        )
        parser.add_argument(
            "--tag",
            action="append",
            default=[],
            dest="tags",
            help=_("Add an extra tag to the report. Can be specified multiple times."),
        )
        parser.add_argument(
            "-v",
            "--version",
            action="store_true",
            help=_("Print the Apport version number."),
        )
        parser.add_argument("issue", nargs="?", help=argparse.SUPPRESS)

        args = parser.parse_args(argv[1:])
        issue = args.issue
        del args.issue

        # mutually exclusive arguments
        if args.update_report:
            args_only_for_new_reports = [
                args.filebug,
                args.window,
                args.symptom,
                args.pid,
                args.crash_file,
                args.save,
            ]
            if any(args_only_for_new_reports):
                parser.error(
                    "-u/--update-bug option cannot be used together"
                    " with options for a new report"
                )

        # no argument: default to "show pending crashes" except when called in
        # bug mode
        # NOTE: uses argv, since args if empty for all the options,
        # e.g. "-v" or "-u $BUG"
        if len(argv) == 1 and cmd.endswith("-bug"):
            args.filebug = True
            return args

        # one argument: guess "file bug" mode by argument type
        if issue is None:
            return args

        # symptom?
        if os.path.exists(os.path.join(symptom_script_dir, issue + ".py")):
            args.filebug = True
            args.symptom = issue

        # .crash/.apport file?
        elif issue.endswith(".crash") or issue.endswith(".apport"):
            args.crash_file = issue

        # PID?
        elif issue.isdigit():
            args.filebug = True
            args.pid = issue

        # executable?
        elif "/" in issue:
            if issue.startswith("/snap/bin"):
                # see if the snap has the same name as the executable
                snap = apport.fileutils.find_snap(issue.split("/")[-1])
                if not snap:
                    parser.error(
                        f"{issue} is provided by a snap. No contact address"
                        f" has been provided; visit the forum at"
                        f" https://forum.snapcraft.io/ for help."
                    )
                elif snap.get("contact", ""):
                    parser.error(
                        f"{issue} is provided by a snap published by"
                        f" {snap['developer']}. Contact them"
                        f" via {snap['contact']} for help."
                    )
                else:
                    parser.error(
                        f"{issue} is provided by a snap published by"
                        f" {snap['developer']}. No contact address"
                        f" has been provided; visit the forum"
                        f" at https://forum.snapcraft.io/ for help."
                    )
                sys.exit(1)
            else:
                pkg = packaging.get_file_package(issue)
                if not pkg:
                    parser.error(f"{issue} does not belong to a package.")
                    sys.exit(1)
            args.filebug = True
            args.package = pkg

        # otherwise: package name
        else:
            args.filebug = True
            self.specified_a_pkg = True
            args.package = issue

        return args

    @staticmethod
    def format_filesize(size):
        """Format the given integer as humanly readable and i18n'ed file
        size."""
        if size < 1000000:
            return locale.format_string("%.1f KB", size / 1000.0)
        if size < 1000000000:
            return locale.format_string("%.1f MB", size / 1000000.0)
        return locale.format_string("%.1f GB", size / float(1000000000))

    def get_complete_size(self):
        """Return the size of the complete report."""
        assert self.report
        # report wasn't loaded, so count manually
        size = 0
        for k in self.report:
            if self.report[k]:
                try:
                    size += self.report[k].get_on_disk_size()
                except AttributeError:
                    size += len(self.report[k])
        return size

    def get_reduced_size(self):
        """Return the size of the reduced report."""
        assert self.report
        size = 0
        for k in self.report:
            if k != "CoreDump":
                if self.report[k]:
                    try:
                        size += self.report[k].get_on_disk_size()
                    except AttributeError:
                        size += len(self.report[k])

        return size

    def can_examine_locally(self):
        """Check whether to offer the "Examine locally" button.

        This will be true if the report has a core dump, apport-retrace is
        installed and a terminal is available (see ui_has_terminal()).
        """
        if not self.report or "CoreDump" not in self.report:
            return False

        if shutil.which("apport-retrace") is None:
            return False

        return self.ui_has_terminal()

    def restart(self):
        """Reopen the crashed application."""
        assert self.report and "ProcCmdline" in self.report

        if os.fork() == 0:
            os.setsid()
            os.execlp(
                "sh",
                "sh",
                "-c",
                self.report.get("RespawnCommand", self.report["ProcCmdline"]),
            )
            sys.exit(1)

    def examine(self):
        """Locally examine crash report."""
        assert self.report_file
        response = self.ui_question_choice(
            _(
                "This will launch apport-retrace in a terminal window"
                " to examine the crash."
            ),
            [
                _("Run gdb session"),
                _("Run gdb session without downloading debug symbols"),
                # TRANSLATORS: %s contains the crash report file name
                _("Update %s with fully symbolic stack trace") % self.report_file,
            ],
            False,
        )

        if response is None:
            return

        cache_dir = os.path.expanduser("~/.cache/apport/retrace")
        retrace_with_download = f"apport-retrace -S system -C {cache_dir} -v "
        retrace_no_download = "apport-retrace "
        filearg = "'" + self.report_file.replace("'", "'\\''") + "'"

        cmds = {
            0: retrace_with_download + "--gdb " + filearg,
            1: retrace_no_download + "--gdb " + filearg,
            2: retrace_with_download + "--output " + filearg + " " + filearg,
        }

        self.ui_run_terminal(cmds[response[0]])

    def remember_send_report(self, send_report: bool) -> None:
        """Put whoopsie in auto or never mode."""
        try:
            subprocess.check_output(
                [
                    "/usr/bin/gdbus",
                    "call",
                    "-y",
                    "-d",
                    "com.ubuntu.WhoopsiePreferences",
                    "-o",
                    "/com/ubuntu/WhoopsiePreferences",
                    "-m",
                    "com.ubuntu.WhoopsiePreferences.SetReportCrashes",
                    str(send_report).lower(),
                ]
            )
            subprocess.check_output(
                [
                    "/usr/bin/gdbus",
                    "call",
                    "-y",
                    "-d",
                    "com.ubuntu.WhoopsiePreferences",
                    "-o",
                    "/com/ubuntu/WhoopsiePreferences",
                    "-m",
                    "com.ubuntu.WhoopsiePreferences.SetAutomaticallyReportCrashes",
                    "true",
                ]
            )
        except (OSError, subprocess.CalledProcessError) as error:
            msg = _(
                "Saving crash reporting state failed."
                " Can't set auto or never reporting mode."
            )
            self.ui_error_message(
                _("Can't remember send report status settings"),
                f"{msg}\n\n{str(error)}",
            )

    def check_report_crashdb(self):
        """Process reports' CrashDB field, if present."""
        if self.report is None or "CrashDB" not in self.report:
            return True

        # specification?
        if self.report["CrashDB"].lstrip().startswith("{"):
            try:
                spec = ast.literal_eval(self.report["CrashDB"])
                assert isinstance(spec, dict)
                assert "impl" in spec
            except (AssertionError, SyntaxError, ValueError) as error:
                self.report["UnreportableReason"] = (
                    f"A package hook defines an invalid crash database"
                    f" definition:\n{self.report['CrashDB']}\n{error}"
                )
                return False
            try:
                self.crashdb = apport.crashdb.load_crashdb(None, spec)
            except (ImportError, KeyError):
                self.report["UnreportableReason"] = (
                    f"A package hook wants to send this report to the crash"
                    f' database "{self.report["CrashDB"]}"'
                    f" which does not exist."
                )

        else:
            # DB name
            try:
                self.crashdb = apport.crashdb.get_crashdb(None, self.report["CrashDB"])
            except (ImportError, KeyError):
                self.report["UnreportableReason"] = (
                    f"A package hook wants to send this report to the crash"
                    f' database "{self.report["CrashDB"]}"'
                    f" which does not exist."
                )
                return False

        return True

    def collect_info(
        self, symptom_script=None, ignore_uninstalled=False, on_finished=None
    ):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        """Collect additional information.

        Call all the add_*_info() methods and display a progress dialog during
        this.

        In particular, this adds OS, package and gdb information and checks bug
        patterns.

        If a symptom script is given, this will be run first (used by
        run_symptom()).
        """
        assert self.report
        self.report["_MarkForUpload"] = "True"

        # skip if we already ran (we might load a processed report)
        if (
            self.report.get("ProblemType") == "Crash" and "Stacktrace" in self.report
        ) or (
            self.report.get("ProblemType") != "Crash" and "Dependencies" in self.report
        ):
            if on_finished:
                on_finished()
            return

        # ensure that the crashed program is still installed:
        if self.report.get("ProblemType") == "Crash":
            exe_path = self.report.get("ExecutablePath", "")
            if not os.path.exists(exe_path):
                msg = _(
                    "This problem report applies to a program"
                    " which is not installed any more."
                )
                if exe_path:
                    msg = f"{msg} ({self.report['ExecutablePath']})"
                self.report["UnreportableReason"] = msg
                if on_finished:
                    on_finished()
                return

            if "InterpreterPath" in self.report:
                if not os.path.exists(self.report["InterpreterPath"]):
                    msg = _(
                        "This problem report applies to a program"
                        " which is not installed any more."
                    )
                    self.report["UnreportableReason"] = (
                        f"{msg} ({self.report['InterpreterPath']})"
                    )
                    if on_finished:
                        on_finished()
                    return

        # check if binary changed since the crash happened
        if "ExecutablePath" in self.report and "ExecutableTimestamp" in self.report:
            orig_time = int(self.report["ExecutableTimestamp"])
            del self.report["ExecutableTimestamp"]
            cur_time = int(os.stat(self.report["ExecutablePath"]).st_mtime)

            if orig_time != cur_time:
                self.report["_MarkForUpload"] = "False"
                self.report["UnreportableReason"] = (
                    _(
                        "The problem happened with the program %s"
                        " which changed since the crash occurred."
                    )
                    % self.report["ExecutablePath"]
                )
                return

        if (
            not self.cur_package
            and "ExecutablePath" not in self.report
            and not symptom_script
        ):
            # this happens if we file a bug without specifying a PID or a
            # package
            self.report.add_os_info()
        else:
            # since this might take a while, create separate threads and
            # display a progress dialog.
            self.ui_start_info_collection_progress()
            # only use a UI for asking questions if the crash db will accept
            # the report
            if self.crashdb.accepts(self.report):
                hookui = HookUI(self)
            else:
                hookui = NoninteractiveHookUI()

            if "Stacktrace" not in self.report:
                # save original environment, in case hooks change it
                orig_env = os.environ.copy()
                icthread = apport.REThread.REThread(
                    target=thread_collect_info,
                    name="thread_collect_info",
                    args=(
                        self.report,
                        self.report_file,
                        self.cur_package,
                        hookui,
                        symptom_script,
                        ignore_uninstalled,
                    ),
                )
                icthread.start()
                while icthread.is_alive():
                    self.ui_pulse_info_collection_progress()
                    try:
                        hookui.process_event()
                    except KeyboardInterrupt:
                        sys.exit(1)

                icthread.join()

                # restore original environment
                os.environ.clear()
                os.environ.update(orig_env)

                try:
                    icthread.exc_raise()
                except (OSError, EOFError, zlib.error) as error:
                    # can happen with broken core dumps
                    msg = _("This problem report is damaged and cannot be processed.")
                    self.report["UnreportableReason"] = f"{msg}\n\n{repr(error)}"
                    self.report["_MarkForUpload"] = "False"
                except ValueError:  # package does not exist
                    if "UnreportableReason" not in self.report:
                        self.report["UnreportableReason"] = _(
                            "This report is about a package that is not installed."
                        )
                        self.report["_MarkForUpload"] = "False"
                except Exception as error:  # pylint: disable=broad-except
                    apport.logging.error("%s", repr(error))
                    self.report["UnreportableReason"] = (
                        _(
                            "An error occurred while attempting to "
                            "process this problem report:"
                        )
                        + "\n\n"
                        + str(error)
                    )
                    self.report["_MarkForUpload"] = "False"

            # ask for target, if snap and deb package are installed
            if "Snap" in self.report and "Package" in self.report:
                if "(not installed)" in self.report["Package"]:
                    # choose snap automatically, if deb package
                    # is not installed
                    res = [0]
                else:
                    res = self.ui_question_choice(
                        _(
                            "You have two versions of this application"
                            " installed, which one do you want to report"
                            " a bug against?"
                        ),
                        [
                            _("%s snap") % self.report["Snap"],
                            _("%s deb package") % self.report["Package"],
                        ],
                        False,
                    )
                # bug report is about the snap, clean deb package info
                if res == [0]:
                    del self.report["Package"]
                    if "PackageArchitecture" in self.report:
                        del self.report["PackageArchitecture"]
                    if "SourcePackage" in self.report:
                        del self.report["SourcePackage"]
                # bug report is about the deb package, clean snap info
                elif res == [1]:
                    del self.report["Snap"]
                    if "SnapSource" in self.report:
                        del self.report["SnapSource"]
                    if "SnapTags" in self.report:
                        del self.report["SnapTags"]
                else:
                    self.ui_stop_info_collection_progress()
                    sys.exit(0)

            # append snap tags, if this report is about the snap
            if "Snap" in self.report and "SnapTags" in self.report:
                self.report.add_tags(self.report.pop("SnapTags").split(" "))

            # show a hint if we cannot auto report a snap bug via 'SnapSource'
            if (
                "Snap" in self.report
                and (
                    "SnapSource" not in self.report and "SnapGitName" not in self.report
                )
                and "UnreportableReason" not in self.report
                and self.specified_a_pkg
            ):
                snap = apport.fileutils.find_snap(self.cur_package)
                if snap.get("contact", ""):
                    self.report["UnreportableReason"] = _(
                        "%s is provided by a snap published by %s."
                        " Contact them via %s for help."
                    ) % (snap["name"], snap["developer"], snap["contact"])
                else:
                    self.report["UnreportableReason"] = _(
                        "%s is provided by a snap published by %s. No contact"
                        " address has been provided; visit the forum"
                        " at https://forum.snapcraft.io/ for help."
                    ) % (snap["name"], snap["developer"])
                self.report["_MarkForUpload"] = "False"

            if "UnreportableReason" in self.report or not self.check_report_crashdb():
                self.ui_stop_info_collection_progress()
                if on_finished:
                    on_finished()
                return

            # check bug patterns
            if (
                self.report.get("ProblemType") == "KernelCrash"
                or self.report.get("ProblemType") == "KernelOops"
                or "Package" in self.report
            ):
                bpthread = apport.REThread.REThread(
                    target=self.report.search_bug_patterns,
                    args=(self.crashdb.get_bugpattern_baseurl(),),
                )
                bpthread.start()
                while bpthread.is_alive():
                    self.ui_pulse_info_collection_progress()
                    try:
                        bpthread.join(0.1)
                    except KeyboardInterrupt:
                        sys.exit(1)
                try:
                    bpthread.exc_raise()
                except (OSError, EOFError, zlib.error) as error:
                    # can happen with broken gz values
                    msg = _("This problem report is damaged and cannot be processed.")
                    self.report["UnreportableReason"] = f"{msg}\n\n{repr(error)}"
                if bpthread.return_value():
                    self.report["_KnownReport"] = bpthread.return_value()

            # check crash database if problem is known
            if self.report.get("ProblemType") != "Bug":
                known_thread = apport.REThread.REThread(
                    target=self.crashdb.known, args=(self.report,)
                )
                known_thread.start()
                while known_thread.is_alive():
                    self.ui_pulse_info_collection_progress()
                    try:
                        known_thread.join(0.1)
                    except KeyboardInterrupt:
                        sys.exit(1)
                known_thread.exc_raise()
                val = known_thread.return_value()
                if val is not None:
                    if val is True:
                        self.report["_KnownReport"] = "1"
                    else:
                        self.report["_KnownReport"] = val

            # anonymize; needs to happen after duplicate checking, otherwise we
            # might damage the stack trace
            anonymize_thread = apport.REThread.REThread(target=self.report.anonymize)
            anonymize_thread.start()
            while anonymize_thread.is_alive():
                self.ui_pulse_info_collection_progress()
                try:
                    anonymize_thread.join(0.1)
                except KeyboardInterrupt:
                    sys.exit(1)
            anonymize_thread.exc_raise()

            self.ui_stop_info_collection_progress()

            # check that we were able to determine package names
            if "UnreportableReason" not in self.report:
                if (
                    (
                        "SourcePackage" not in self.report
                        and "Dependencies" not in self.report
                    )
                    or (
                        not self.report.get("ProblemType", "").startswith("Kernel")
                        and "Package" not in self.report
                    )
                ) and not self._is_snap():
                    self.ui_error_message(
                        _("Invalid problem report"),
                        _("Could not determine the package or source package name."),
                    )
                    # TODO This is not called consistently,
                    # is it really needed?
                    self.ui_shutdown()
                    sys.exit(1)

        if on_finished:
            on_finished()

    def _is_snap(self):
        assert self.report
        return "SnapSource" in self.report or "Snap" in self.report

    def open_url(self, url):
        """Open the given URL in a new browser window.

        Display an error dialog if everything fails.
        """
        try:
            try:
                run_as_real_user(["xdg-open", url], get_user_env=True)
                return
            except OSError:
                # fall back to webbrowser
                if webbrowser.open(url, new=1, autoraise=True):
                    return
                error_details = ""
        except Exception as error:  # pylint: disable=broad-except
            error_details = f"\n{error}"

        title = _("Unable to start web browser")
        message = _("Unable to start web browser to open %s.") % url
        self.ui_error_message(title, message + error_details)

    def file_report(self):
        """Upload the current report and guide the user to the reporting
        web page."""
        assert self.report
        # FIXME: This behaviour is not really correct, but necessary as
        # long as we only support a single crashdb and have whoopsie
        # hardcoded. Once we have multiple crash dbs, we need to check
        # accepts() earlier, and not even present the data if none of
        # the DBs wants the report. See LP#957177 for details.
        if not self.crashdb.accepts(self.report):
            return
        # drop PackageArchitecture if equal to Architecture
        if self.report.get("PackageArchitecture") == self.report.get("Architecture"):
            try:
                del self.report["PackageArchitecture"]
            except KeyError:
                pass

        # StacktraceAddressSignature is redundant and does not need to clutter
        # the database
        try:
            del self.report["StacktraceAddressSignature"]
        except KeyError:
            pass

        self.upload_progress = None

        def progress_callback(sent, total):
            self.upload_progress = float(sent) / total

        message_queue = queue.SimpleQueue()

        def message_callback(title, text):
            message_displayed = threading.Event()
            message_queue.put((title, text, message_displayed))
            message_displayed.wait()

        # drop internal/uninteresting keys, that start with "_"
        for k in list(self.report):
            if k.startswith("_"):
                del self.report[k]

        self.ui_start_upload_progress()
        upthread = apport.REThread.REThread(
            target=self.crashdb.upload,
            args=(self.report, progress_callback, message_callback),
        )
        upthread.start()
        try:
            while upthread.is_alive():
                self.ui_set_upload_progress(self.upload_progress)
                try:
                    title, text, msg_displayed = message_queue.get(
                        block=True, timeout=0.1
                    )
                    self.ui_info_message(title, text)
                    msg_displayed.set()
                    upthread.exc_raise()
                except queue.Empty:
                    pass

            upthread.exc_raise()
        except KeyboardInterrupt:
            sys.exit(1)
        except (smtplib.SMTPConnectError, urllib.error.URLError) as error:
            msg = _(
                "Cannot connect to crash database,"
                " please check your Internet connection."
            )
            self.ui_error_message(_("Network problem"), f"{msg}\n\n{str(error)}")
            return

        ticket = upthread.return_value()
        self.ui_stop_upload_progress()

        url = self.crashdb.get_comment_url(self.report, ticket)
        if url:
            self.open_url(url)

    def load_report(self, path):
        """Load report from given path and do some consistency checks.

        This might issue an error message and return False if the report cannot
        be processed, otherwise self.report is initialized and True is
        returned.
        """
        try:
            self.report = apport.Report()
            with open(path, "rb") as f:
                self.report.load(f, binary="compressed")
            if "ProblemType" not in self.report:
                raise ValueError('Report does not contain "ProblemType" field')
        except MemoryError:
            self.report = None
            self.ui_error_message(
                _("Memory exhaustion"),
                _(
                    "Your system does not have enough memory"
                    " to process this crash report."
                ),
            )
            return False
        except OSError as error:
            self.report = None
            self.ui_error_message(_("Invalid problem report"), error.strerror)
            return False
        except (TypeError, ValueError, AssertionError, zlib.error) as error:
            self.report = None
            msg = _("This problem report is damaged and cannot be processed.")
            self.ui_error_message(
                _("Invalid problem report"), f"{msg}\n\n{repr(error)}"
            )
            return False

        if "Package" in self.report:
            self.cur_package = self.report["Package"].split()[0]
        else:
            self.cur_package = apport.fileutils.find_file_package(
                self.report.get("ExecutablePath", "")
            )

        return True

    def check_unreportable(self):
        """Check if the current report is unreportable.

        If so, display an info message and return True.
        """
        assert self.report
        if not self.crashdb.accepts(self.report):
            return False
        if "UnreportableReason" in self.report:
            if isinstance(self.report["UnreportableReason"], bytes):
                self.report["UnreportableReason"] = self.report[
                    "UnreportableReason"
                ].decode("UTF-8")
            if "Package" in self.report:
                title = _("Problem in %s") % self.report["Package"].split()[0]
            else:
                title = ""
            self.ui_info_message(
                title,
                _("The problem cannot be reported:\n\n%s")
                % self.report["UnreportableReason"],
            )
            return True
        return False

    def get_desktop_entry(self):
        """Return a .desktop info dictionary for the current report.

        Return None if report cannot be associated to a .desktop file.
        """
        assert self.report
        if "DesktopFile" in self.report and os.path.exists(self.report["DesktopFile"]):
            desktop_file = self.report["DesktopFile"]
        else:
            try:
                desktop_file = apport.fileutils.find_package_desktopfile(
                    self.cur_package
                )
            except ValueError:
                return None

        if not desktop_file:
            return None

        cp = configparser.ConfigParser(interpolation=None, strict=False)
        try:
            cp.read(desktop_file, encoding="UTF-8")
        except configparser.Error as error:
            sys.stderr.write(f"Warning! {desktop_file} is broken: {str(error)}\n")
            return None
        if not cp.has_section("Desktop Entry"):
            return None
        result = dict(cp.items("Desktop Entry"))
        if "name" not in result:
            return None
        return result

    def handle_duplicate(self):
        """Check if current report matches a bug pattern.

        If so, tell the user about it, open the existing bug in a browser, and
        return True.
        """
        assert self.report
        if not self.crashdb.accepts(self.report):
            return False
        if "_KnownReport" not in self.report:
            return False

        # if we have an URL, open it; otherwise this is just a marker that we
        # know about it
        if self.report["_KnownReport"].startswith("http"):
            self.ui_info_message(
                _("Problem already known"),
                _(
                    "This problem was already reported in the bug report"
                    " displayed in the web browser. Please check"
                    " if you can add any further information"
                    " that might be helpful for the developers."
                ),
            )

            self.open_url(self.report["_KnownReport"])
        else:
            self.ui_info_message(
                _("Problem already known"),
                _("This problem was already reported to developers. Thank you!"),
            )

        return True

    def add_extra_tags(self):
        """Add extra tags to report specified with --tags on CLI."""
        assert self.report
        if self.args.tags:
            self.report.add_tags(self.args.tags)

    #
    # abstract UI methods that must be implemented in derived classes
    #

    def ui_present_report_details(
        self, allowed_to_report: bool = True, modal_for: int | None = None
    ) -> Action:
        """Show details of the bug report.

        Return the action and options as an Action object:

        - Valid attributes are: report the crash ('report'), restart
          the crashed application ('restart'), or ignore further crashes
          ('ignore').
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_info_message(self, title, text):
        """Show an information message box with given title and text."""
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_error_message(self, title, text):
        """Show an error message box with given title and text."""
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_start_info_collection_progress(self):
        """Open a indefinite progress bar for data collection.

        This tells the user to wait while debug information is being
        collected.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_pulse_info_collection_progress(self):
        """Advance the data collection progress bar.

        This function is called every 100 ms.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_stop_info_collection_progress(self):
        """Close debug data collection progress window."""
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_start_upload_progress(self):
        """Open progress bar for data upload.

        This tells the user to wait while debug information is being uploaded.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_set_upload_progress(self, progress: float | None) -> None:
        """Update data upload progress bar.

        Set the progress bar in the debug data upload progress window to the
        given ratio (between 0 and 1, or None for indefinite progress).

        This function is called every 100 ms.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_stop_upload_progress(self):
        """Close debug data upload progress window."""
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_shutdown(self):
        """Called right before terminating the program.

        This can be used for for cleaning up.
        """

    def ui_has_terminal(self):
        """Check for a terminal window.

        Check if a terminal application is available and can be launched.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_run_terminal(self, command):
        """Run command in a terminal window.

        Run given command in a terminal window; raise an
        exception if terminal cannot be opened.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    #
    # Additional UI dialogs; these are not required by Apport itself, but can
    # be used by interactive package hooks
    #

    def ui_question_yesno(self, text):
        """Show a yes/no question.

        Return True if the user selected "Yes", False if selected "No" or
        "None" on cancel/dialog closing.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_question_choice(self, text, options, multiple):
        """Show an question with predefined choices.

        options is a list of strings to present. If multiple is True, they
        should be check boxes, if multiple is False they should be radio
        buttons.

        Return list of selected option indexes, or None if the user cancelled.
        If multiple == False, the list will always have one element.
        """
        raise NotImplementedError("this function must be overridden by subclasses")

    def ui_question_file(self, text):
        """Show a file selector dialog.

        Return path if the user selected a file, or None if cancelled.
        """
        raise NotImplementedError("this function must be overridden by subclasses")


class HookUI:
    """Interactive functions which can be used in package hooks.

    This provides an interface for package hooks which need to ask interactive
    questions. Directly passing the UserInterface instance to the hooks needs
    to be avoided, since we need to call the UI methods in a different thread,
    and also don't want hooks to be able to poke in the UI.
    """

    def __init__(self, ui):
        """Create a HookUI object.

        ui is the UserInterface instance to wrap.
        """
        self.ui = ui

        # variables for communicating with the UI thread
        self._request_event = threading.Event()
        self._response_event = threading.Event()
        self._request_fn = None
        self._request_args = None
        self._response = None

    #
    # API for hooks
    #

    def information(self, text):
        """Show an information with OK/Cancel buttons.

        This can be used for asking the user to perform a particular action,
        such as plugging in a device which does not work.
        """
        return self._trigger_ui_request("ui_info_message", "", text)

    def yesno(self, text):
        """Show a yes/no question.

        Return True if the user selected "Yes", False if selected "No" or
        "None" on cancel/dialog closing.
        """
        return self._trigger_ui_request("ui_question_yesno", text)

    def choice(self, text, options, multiple=False):
        """Show an question with predefined choices.

        options is a list of strings to present. If multiple is True, they
        should be check boxes, if multiple is False they should be radio
        buttons.

        Return list of selected option indexes, or None if the user cancelled.
        If multiple == False, the list will always have one element.
        """
        return self._trigger_ui_request("ui_question_choice", text, options, multiple)

    def file(self, text):
        """Show a file selector dialog.

        Return path if the user selected a file, or None if cancelled.
        """
        return self._trigger_ui_request("ui_question_file", text)

    #
    # internal API for inter-thread communication
    #

    def _trigger_ui_request(self, fn, *args):
        """Called by HookUi functions in info collection thread."""
        # only one at a time
        assert not self._request_event.is_set()
        assert not self._response_event.is_set()
        assert self._request_fn is None

        self._response = None
        self._request_fn = fn
        self._request_args = args
        self._request_event.set()
        self._response_event.wait()

        self._request_fn = None
        self._response_event.clear()

        return self._response

    def process_event(self):
        """Called by GUI thread to check and process hook UI requests."""
        # sleep for 0.1 seconds to wait for events
        self._request_event.wait(0.1)
        if not self._request_event.is_set():
            return

        assert not self._response_event.is_set()
        self._request_event.clear()
        self._response = getattr(self.ui, self._request_fn)(*self._request_args)
        self._response_event.set()


class NoninteractiveHookUI(HookUI):
    """HookUI variant that does not ask the user any questions."""

    def __init__(self):
        super().__init__(None)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def information(self, text):
        return None

    def yesno(self, text):
        return None

    def choice(self, text, options, multiple=False):
        return None

    def file(self, text):
        return None

    def process_event(self):
        # Give other threads some chance to run
        time.sleep(0.1)



"""System tests for apport-retrace."""

import os
import pathlib
import re
import shutil
import signal
import subprocess
import tempfile
import textwrap
from collections.abc import Iterator

import pytest

from apport.packaging_impl.apt_dpkg import impl
from apport.report import Report
from tests.paths import get_test_data_directory, local_test_environment

CODENAME_DISTRO_RELEASE_MAP = {"jammy": "Ubuntu 22.04"}


@pytest.fixture(name="module_workdir", scope="module")
def fixture_module_workdir() -> Iterator[pathlib.Path]:
    """Create a temporary work directory for all test case in this module."""
    workdir = tempfile.mkdtemp(prefix="apport_retrace_system_tests_", dir="/var/tmp")
    yield pathlib.Path(workdir)
    shutil.rmtree(workdir)


@pytest.fixture(name="module_cachedir", scope="module")
def fixture_module_cachedir(module_workdir: pathlib.Path) -> pathlib.Path:
    """Return sandbox cache directory used for all test in this module."""
    return module_workdir / "cache"


@pytest.fixture(name="divide_by_zero_crash", scope="module")
def fixture_divide_by_zero_crash(module_workdir: pathlib.Path) -> str:
    """Generate a Report with a crash of divide-by-zero."""
    executable = "/usr/bin/divide-by-zero"
    assert pathlib.Path(executable).exists()
    core_path = module_workdir / "divide-by-zero.core"
    gdb = subprocess.run(
        [
            "gdb",
            "--batch",
            "-iex",
            "set debuginfod enable off",
            "--ex",
            "run",
            "--ex",
            f"generate-core-file {core_path}",
            executable,
        ],
        check=True,
        stdout=subprocess.PIPE,
        text=True,
    )
    assert core_path.exists()
    signal_match = re.search("Program received signal (SIG[A-Z]+)", gdb.stdout)
    assert signal_match, gdb.stdout

    # generate crash report
    report = Report()
    report["ExecutablePath"] = executable
    report["Signal"] = str(signal.Signals[signal_match.group(1)].value)
    report["SignalName"] = signal_match.group(1)
    report.add_os_info()
    report.add_package_info()
    report["CoreDump"] = (str(core_path),)

    report_filename = module_workdir / "divide-by-zero.crash"
    with open(report_filename, "wb") as report_file:
        report.write(report_file)

    return str(report_filename)


@pytest.fixture(name="workdir")
def fixture_workdir() -> Iterator[pathlib.Path]:
    """Create a temporary work directory for the test case."""
    workdir = tempfile.mkdtemp(prefix="apport_retrace_system_test_", dir="/var/tmp")
    yield pathlib.Path(workdir)
    shutil.rmtree(workdir)


def setup_ubuntu_sandbox_config(
    configdir: pathlib.Path, release: str, arch: str
) -> None:
    """Set up sandbox configuration for retracing Ubuntu crashes."""
    config_release_dir = configdir / CODENAME_DISTRO_RELEASE_MAP[release]
    codename_file = config_release_dir / "codename"
    sources_dir = config_release_dir / "sources.list.d"
    sources_file = sources_dir / "ubuntu.sources"
    # pylint: disable-next=protected-access
    uri = impl._get_mirror(arch)

    sources_dir.mkdir(parents=True)
    codename_file.write_text(release)
    sources_file.write_text(
        textwrap.dedent(
            f"""\
            Types: deb deb-src
            URIs: {uri}
            Suites: {release}
            Architertures: {arch}
            Components: main
            Signed-By: /usr/share/keyrings/ubuntu-archive-keyring.gpg
            """
        )
    )


@pytest.fixture(name="sandbox_config")
def fixture_sandbox_config(workdir: pathlib.Path) -> str:
    """Setup a sandbox configuration that supports Ubuntu jammy."""
    config = workdir / "config"
    config.mkdir()
    setup_ubuntu_sandbox_config(config, "jammy", "amd64")
    return str(config)


def _read_and_print_retraced_report(report_filename: pathlib.Path) -> Report:
    report = Report()
    with open(report_filename, "rb") as report_file:
        report.load(report_file)

    assert "CoreDump" in report
    del report["CoreDump"]
    print(f"Retraced report without CoreDump: {report}")

    return report


def _assert_is_retraced(report: Report) -> None:
    for expected_key in (
        "Disassembly",
        "Registers",
        "Stacktrace",
        "StacktraceSource",
        "StacktraceTop",
        "ThreadStacktrace",
    ):
        assert expected_key in report


def _assert_divide_by_zero_retrace(report: Report) -> None:
    # The path could be relative to the source code (with or without
    # a leading "./" or it can be absolute with /usr/src/<deb-source>-<deb-version>

    # noble: at /usr/src/chaos-marmosets-0.2.0-1/divide-by-zero.c
    # noble armhf: at divide-by-zero.c
    # jammy: at ./divide-by-zero.c

    # Actual frame info format depends on arch, compiler, opt-level...
    # If the failing instruction is *not* the first instruction for the
    # line there's an addition pointer in the line.
    frame_regex = re.compile(
        r"^#0  (0x[0-9a-f]+ in )?divide_by_zero \(\)"
        r" at (/usr/src/chaos-marmosets-[^/]+/|\./)?divide-by-zero.c:[0-9]+$",
        flags=re.M,
    )
    stack_regex = re.compile(
        r"^divide_by_zero \(\)"
        r" at (/usr/src/chaos-marmosets-[^/]+/|\./)?divide-by-zero.c:[0-9]+$",
        flags=re.M,
    )
    assert "divide_by_zero" in report["Disassembly"]
    # Expect RIP point to divide_by_zero
    assert "divide_by_zero" in report["Registers"]
    assert frame_regex.match(report["Stacktrace"])
    assert frame_regex.match(report["StacktraceSource"])
    assert "42 / zero" in report["StacktraceSource"]
    assert stack_regex.match(report["StacktraceTop"])
    assert frame_regex.search(report["ThreadStacktrace"])


def _assert_sleep_retrace(report: Report) -> None:
    stack_top = " in __GI___clock_nanosleep "
    assert "__GI___clock_nanosleep" in report["Disassembly"]
    # Expect RIP point to offset of __GI___clock_nanosleep
    assert "__GI___clock_nanosleep" in report["Registers"]
    assert stack_top in report["Stacktrace"]
    assert "seconds = 86400" in report["Stacktrace"]
    assert stack_top in report["StacktraceSource"]
    assert "return nanosleep" in report["StacktraceSource"]
    assert "__GI___clock_nanosleep (clock_id=" in report["StacktraceTop"]
    assert stack_top in report["ThreadStacktrace"]
    assert "seconds = 86400" in report["ThreadStacktrace"]


def _assert_cache_has_content(
    cachedir: pathlib.Path, architecture: str, codename: str
) -> None:
    # pylint: disable=protected-access
    release = CODENAME_DISTRO_RELEASE_MAP[codename]
    aptroot = pathlib.Path(impl._apt_cache_root_dir(architecture, cachedir, release))
    assert list((aptroot / "var/lib/apt/lists").iterdir())
    assert list((aptroot / "var/cache/apt/archives").iterdir())


@pytest.mark.requires_internet
@pytest.mark.skipif(
    impl.get_system_architecture() == "s390x",
    reason="GDB has issues with divide-by-zero on s390x (LP: #2075204)",
)
def test_retrace_system_sandbox(
    workdir: pathlib.Path, module_cachedir: pathlib.Path, divide_by_zero_crash: str
) -> None:
    """Retrace a divide-by-zero crash in a system sandbox."""
    retraced_report_filename = workdir / "retraced.crash"
    env = os.environ | local_test_environment()
    cmd = [
        "apport-retrace",
        "-v",
        "-o",
        str(retraced_report_filename),
        "--sandbox",
        "system",
        "--cache",
        str(module_cachedir),
        divide_by_zero_crash,
    ]
    subprocess.run(cmd, check=True, env=env)

    report = _read_and_print_retraced_report(retraced_report_filename)
    _assert_is_retraced(report)
    _assert_divide_by_zero_retrace(report)


@pytest.mark.requires_internet
@pytest.mark.skipif(
    impl.get_system_architecture() != "amd64",
    reason="Testing the GDB sandbox erroring out on non-AMD64",
)
def test_retrace_system_sandbox_gdb_sandbox(
    workdir: pathlib.Path, module_cachedir: pathlib.Path, divide_by_zero_crash: str
) -> None:
    """Retrace a divide-by-zero crash in a system sandbox with a GDB sandbox."""
    retraced_report_filename = workdir / "retraced.crash"
    env = os.environ | local_test_environment()
    cmd = [
        "apport-retrace",
        "-v",
        "-o",
        str(retraced_report_filename),
        "--sandbox",
        "system",
        "--gdb-sandbox",
        "--cache",
        str(module_cachedir),
        divide_by_zero_crash,
    ]
    subprocess.run(cmd, check=True, env=env)

    report = _read_and_print_retraced_report(retraced_report_filename)
    _assert_is_retraced(report)
    _assert_divide_by_zero_retrace(report)


@pytest.mark.requires_internet
@pytest.mark.skipif(
    impl.get_system_architecture() != "amd64" and shutil.which("gdb-multiarch") is None,
    reason="gdb-multiarch is needed for proper retracing on foreign architectures",
)
def test_retrace_jammy_sandbox(
    workdir: pathlib.Path, module_cachedir: pathlib.Path, sandbox_config: pathlib.Path
) -> None:
    """Retrace a sleep crash from jammy in a sandbox."""
    crash = get_test_data_directory() / "jammy_usr_bin_sleep.1000.crash"
    retraced_report_filename = workdir / "retraced.crash"
    env = os.environ | local_test_environment()
    cmd = [
        "apport-retrace",
        "-v",
        "-o",
        str(retraced_report_filename),
        "--sandbox",
        str(sandbox_config),
        "--cache",
        str(module_cachedir),
        "--sandbox-dir",
        str(workdir / "apport_sandbox"),
        str(crash),
    ]
    subprocess.run(cmd, check=True, env=env)

    report = _read_and_print_retraced_report(retraced_report_filename)
    _assert_is_retraced(report)
    _assert_sleep_retrace(report)
    _assert_cache_has_content(module_cachedir, "amd64", "jammy")


@pytest.mark.requires_internet
@pytest.mark.skipif(
    impl.get_system_architecture() != "amd64",
    reason="GDB sandbox only available on amd64",
)
def test_retrace_jammy_sandbox_gdb_sandbox(
    workdir: pathlib.Path, module_cachedir: pathlib.Path, sandbox_config: pathlib.Path
) -> None:
    """Retrace a sleep crash from jammy in a sandbox with a GDB sandbox."""
    crash = get_test_data_directory() / "jammy_usr_bin_sleep.1000.crash"
    retraced_report_filename = workdir / "retraced.crash"
    env = os.environ | local_test_environment()
    cmd = [
        "apport-retrace",
        "-v",
        "-o",
        str(retraced_report_filename),
        "--sandbox",
        str(sandbox_config),
        "--gdb-sandbox",
        "--cache",
        str(module_cachedir),
        "--sandbox-dir",
        str(workdir / "apport_sandbox"),
        str(crash),
    ]
    subprocess.run(cmd, check=True, env=env)

    report = _read_and_print_retraced_report(retraced_report_filename)
    _assert_is_retraced(report)
    _assert_sleep_retrace(report)
    _assert_cache_has_content(module_cachedir, "amd64", "jammy")



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.packaging_impl.apt_dpkg."""

import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apt

from apport.packaging_impl.apt_dpkg import (
    WITH_DEB822_SUPPORT,
    _map_mirror_to_arch,
    _parse_deb822_sources,
    _read_mirror_file,
    impl,
)


@unittest.mock.patch(
    "apport.packaging_impl.apt_dpkg.__AptDpkgPackageInfo.get_os_version",
    MagicMock(return_value=("Ubuntu", "22.04")),
)
class TestPackagingAptDpkg(unittest.TestCase):
    """Unit tests for apport.packaging_impl.apt_dpkg."""

    maxDiff = None

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_candidate(self, apt_cache_mock):
        """is_distro_package() for package that has no candidate."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, installed=None, candidate=None
        )
        self.assertEqual(impl.is_distro_package("adduser"), False)
        getitem_mock.assert_called_once_with("adduser")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_installed_version(self, apt_cache_mock):
        """is_distro_package() for not installed package."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package,
            installed=MagicMock(spec=apt.package.Version, version=None),
        )
        self.assertEqual(impl.is_distro_package("7zip"), False)
        getitem_mock.assert_called_once_with("7zip")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_ppa(self, apt_cache_mock):
        """is_distro_package() for a PPA package."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="LP-PPA")],
            version="0.5.0-0ppa1",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        self.assertEqual(impl.is_distro_package("bdebstrap"), False)
        getitem_mock.assert_called_once_with("bdebstrap")

    @unittest.mock.patch("os.path.exists")
    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_system_image(self, apt_cache_mock, exists_mock):
        """is_distro_package() for a system image without cache."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="")],
            version="2.23.1-0ubuntu4",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        exists_mock.return_value = True
        self.assertEqual(impl.is_distro_package("apport"), True)
        getitem_mock.assert_called_once_with("apport")
        exists_mock.assert_called_once_with("/etc/system-image/channel.ini")

    def test_get_library_paths_unknown_arch(self) -> None:
        """get_library_paths() for unknown architecture."""
        with self.assertRaisesRegex(NotImplementedError, "architecture osprey64"):
            impl.get_library_paths("osprey64")

    def test_map_mirror_to_arch_ports_to_primary(self) -> None:
        """Test _map_mirror_to_arch() to map ports to primary."""
        self.assertEqual(
            _map_mirror_to_arch("http://ports.ubuntu.com/ubuntu-ports/", "amd64"),
            "http://archive.ubuntu.com/ubuntu",
        )

    def test_map_mirror_to_arch_country_ports_to_primary(self) -> None:
        """Test _map_mirror_to_arch() to map country ports to primary."""
        self.assertEqual(
            _map_mirror_to_arch("http://de.ports.ubuntu.com/ubuntu-ports", "amd64"),
            "http://de.archive.ubuntu.com/ubuntu",
        )

    def test_map_mirror_to_arch_ports_unchanged(self) -> None:
        """Test _map_mirror_to_arch() to keep ports unchanged."""
        uri = "http://de.ports.ubuntu.com/ubuntu-ports"
        self.assertEqual(_map_mirror_to_arch(uri, "ppc64el"), uri)

    def test_map_mirror_to_arch_primary_to_ports(self) -> None:
        """Test _map_mirror_to_arch() to map primary to ports."""
        self.assertEqual(
            _map_mirror_to_arch("http://de.archive.ubuntu.com/ubuntu/", "s390x"),
            "http://de.ports.ubuntu.com/ubuntu-ports",
        )

    def test_map_mirror_to_arch_primary_unchanged(self) -> None:
        """Test _map_mirror_to_arch() to keep ports unchanged."""
        uri = "http://de.archive.ubuntu.com/ubuntu"
        self.assertEqual(_map_mirror_to_arch(uri, "amd64"), uri)

    @unittest.skipIf(not WITH_DEB822_SUPPORT, "no deb822 support")
    @unittest.mock.patch(
        "builtins.open",
        new_callable=unittest.mock.mock_open,
        read_data="""
# Some documentation in the beginning

Types: deb deb-src
URIs: http://example.com
Suites: foo foo-bar
Components: main


Types: deb
URIs: http://example2.com
Suites:
 baz
Components: main


""",
    )
    def test_parse_deb822_sources_extra_lines(self, mock_file):
        """Test _parse_deb822_sources with multiple lines separating blocks."""
        entries = _parse_deb822_sources("foo_bar.sources")
        mock_file.assert_called_with("foo_bar.sources", encoding="utf-8")
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0].uris[0], "http://example.com")
        self.assertEqual(entries[1].suites[0], "baz")

    def test_read_mirror_file(self) -> None:
        """Test _read_mirror_file with config from GitHub CI."""
        with tempfile.NamedTemporaryFile("w") as mirror_file:
            mirror_file.write(
                "http://azure.archive.ubuntu.com/ubuntu/\tpriority:1\n"
                "http://archive.ubuntu.com/ubuntu/\tpriority:2\n"
                "http://security.ubuntu.com/ubuntu/\tpriority:3\n"
            )
            mirror_file.flush()
            mirrors = _read_mirror_file(f"mirror+file:{mirror_file.name}")
        self.assertEqual(
            mirrors,
            [
                "http://azure.archive.ubuntu.com/ubuntu/",
                "http://archive.ubuntu.com/ubuntu/",
                "http://security.ubuntu.com/ubuntu/",
            ],
        )

    @unittest.mock.patch.object(impl, "_get_file2pkg_mapping")
    @unittest.mock.patch.object(impl, "_save_contents_mapping", MagicMock())
    def test_get_file_package_uninstalled_usrmerge(
        self, _get_file2pkg_mapping_mock: MagicMock
    ) -> None:
        """get_file_package() on uninstalled usrmerge packages."""
        # Data from Ubuntu 24.04 (noble)
        _get_file2pkg_mapping_mock.return_value = {
            b"usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2": b"libc6",
            b"usr/lib/x86_64-linux-gnu/libc.so.6": b"libc6",
            b"usr/libx32/libc.so.6": b"libc6-x32",
        }

        pkg = impl.get_file_package(
            "/lib/x86_64-linux-gnu/libc.so.6", True, "/map_cachedir", arch="amd64"
        )

        self.assertEqual(pkg, "libc6")
        _get_file2pkg_mapping_mock.assert_called_with(
            "/map_cachedir", impl.get_distro_codename(), "amd64"
        )

    def test_contents_path_filering(self) -> None:
        """Test _update_given_file2pkg_mapping to ignore unrelevant files."""
        # Test content taken from
        # http://archive.ubuntu.com/ubuntu/dists/noble/Contents-amd64.gz
        contents = b"""\
bin/ip							    net/iproute2
boot/ipxe.efi						    admin/grub-ipxe
etc/dput.cf						    devel/dput
lib/modules/6.8.0-31-generic/kernel/crypto/lz4.ko.zst	\
    kernel/linux-modules-6.8.0-31-generic
lib/nut/clone						    admin/nut-server
sbin/hdparm						    admin/hdparm
usr/Brother/inf/braddprinter				    multiverse/text/brother-lpr-drivers-laser
usr/aarch64-linux-gnu/include/ar.h			    libdevel/libc6-dev-arm64-cross
usr/bin/git						    vcs/git
usr/bin/xz						    utils/xz-utils
usr/games/etr						    universe/games/extremetuxracer
usr/include/apache2/os.h				    httpd/apache2-dev
usr/include/x86_64-linux-gnu/bits/endian.h		    libdevel/libc6-dev
usr/lib/7zip/7z.so					    universe/utils/7zip
usr/lib/debug/.build-id/31/1c9c9b30d6991fb903ab459173c66eb8e7e895.debug debug/libc6-dbg
usr/libexec/coreutils/libstdbuf.so			    utils/coreutils
usr/libx32/ld.so					    libs/libc6-x32
usr/sbin/zic						    libs/libc-bin
usr/share/cargo/registry/libc-0.2.152/src/lib.rs	    universe/rust/librust-libc-dev
usr/share/dicom3tools/gen.so				    universe/graphics/dicom3tools
usr/share/doc/0install					    universe/admin/0install
usr/share/gimp/2.0/help/de/glossary.html		    universe/doc/gimp-help-de
usr/share/gocode/src/launchpad.net/mgo			    universe/devel/golang-gopkg-mgo.v2-dev
usr/share/help/C/eog/default.page			    gnome/eog
usr/share/icons/gnome-colors-common/32x32/apps/konsole.png\
  universe/gnome/gnome-colors-common
usr/share/libreoffice/help/en-US/noscript.html		    doc/libreoffice-help-en-us
usr/share/locale/de/LC_MESSAGES/apt.mo			    admin/apt
usr/share/man/de/man1/man.1.gz				    doc/man-db
usr/share/texlive/index.html				    universe/tex/texlive-base
usr/src/broadcom-sta.tar.xz				    restricted/admin/broadcom-sta-source
var/lib/ieee-data/iab.txt				    net/ieee-data
"""

        file2pkg: dict[bytes, bytes] = {}
        open_mock = unittest.mock.mock_open(read_data=contents)
        with unittest.mock.patch("gzip.open", open_mock):
            # pylint: disable-next=protected-access
            impl._update_given_file2pkg_mapping(file2pkg, "Contents-amd64", "noble")

        self.assertEqual(
            {k.decode(): v.decode() for k, v in file2pkg.items()},
            {
                "bin/ip": "iproute2",
                "boot/ipxe.efi": "grub-ipxe",
                "etc/dput.cf": "dput",
                "lib/nut/clone": "nut-server",
                "sbin/hdparm": "hdparm",
                "usr/Brother/inf/braddprinter": "brother-lpr-drivers-laser",
                "usr/bin/git": "git",
                "usr/lib/debug/.build-id/31/1c9c9b30d6991fb903ab459173c66eb8e7e895"
                ".debug": "libc6-dbg",
                "usr/bin/xz": "xz-utils",
                "usr/games/etr": "extremetuxracer",
                "usr/lib/7zip/7z.so": "7zip",
                "usr/libexec/coreutils/libstdbuf.so": "coreutils",
                "usr/libx32/ld.so": "libc6-x32",
                "usr/sbin/zic": "libc-bin",
                "usr/share/dicom3tools/gen.so": "dicom3tools",
                "var/lib/ieee-data/iab.txt": "ieee-data",
            },
        )
        open_mock.assert_called_once_with("Contents-amd64", "rb")



"""Legacy logging functions."""

import os
import sys
import time
import typing


def log(message, timestamp=False):
    """Log the given string to stdout. Prepend timestamp if requested."""
    if timestamp:
        sys.stdout.write(f"{time.strftime('%x %X')}: ")
    print(message)


def fatal(msg: str, *args: typing.Any) -> typing.NoReturn:
    """Print out an error message and exit the program."""
    error(msg, *args)
    sys.exit(1)


def error(msg, *args):
    """Print out an error message."""
    if sys.stderr:
        sys.stderr.write("ERROR: ")
        sys.stderr.write(msg % args)
        sys.stderr.write("\n")


def warning(msg, *args):
    """Print out an warning message."""
    if sys.stderr:
        sys.stderr.write("WARNING: ")
        sys.stderr.write(msg % args)
        sys.stderr.write("\n")


def memdbg(checkpoint):
    """Print current memory usage.

    This is only done if $APPORT_MEMDEBUG is set.
    """
    if "APPORT_MEMDEBUG" not in os.environ or not sys.stderr:
        return

    memstat = {}
    with open("/proc/self/status", encoding="utf-8") as status_file:
        for line in status_file:
            if line.startswith("Vm"):
                (field, size, _) = line.split()
                memstat[field[:-1]] = int(size) / 1024.0

    sys.stderr.write(
        f"Size: {memstat['VmSize']:.1f} MB, RSS: {memstat['VmRSS']:.1f} MB,"
        f" Stk: {memstat['VmStk']:.1f} MB @ {checkpoint}\n"
    )



"""Platform-specific apport.packaging implementation."""

import importlib
import os
import platform

from apport.packaging import PackageInfo


def determine_packaging_implementation() -> str:
    """Determine the packaging implementation for the host."""
    info = platform.freedesktop_os_release()
    assert info is not None
    ids = set([info["ID"]]) | set(info.get("ID_LIKE", "").split(" "))
    if "debian" in ids:
        return "apt_dpkg"
    if os.path.exists("/usr/bin/rpm"):
        return "rpm"
    raise RuntimeError(
        "Could not determine system package manager."
        " Please file a bug and provide /etc/os-release!"
    )


def load_packaging_implementation() -> PackageInfo:
    """Return the packaging implementation for the host."""
    module = importlib.import_module(
        f"apport.packaging_impl.{determine_packaging_implementation()}"
    )
    return module.impl


impl = load_packaging_implementation()



"""Setuptools extension to build the Java subdirectory."""

import functools
import glob
import logging
import os
import pathlib
import subprocess
import typing

from setuptools import Command, Distribution


# pylint: disable-next=invalid-name
class build_java(Command):
    """Compile Java components of Apport"""

    description = __doc__
    user_options = [("minimum-java-release=", "r", "Specify minimum Java release.")]

    def __init__(self, dist: Distribution, **kwargs: dict[str, typing.Any]) -> None:
        Command.__init__(self, dist, **kwargs)
        self.initialize_options()

    def initialize_options(self) -> None:
        """Set or (reset) all options/attributes/caches to their default values"""
        self.minimum_java_release = "7"

    def finalize_options(self) -> None:
        """Set final values for all options/attributes"""

    # pylint: disable-next=no-self-use
    def run(self) -> None:
        """Build the Java .class and .jar files."""
        oldwd = os.getcwd()
        os.chdir("java")
        javac = [
            "javac",
            "-source",
            self.minimum_java_release,
            "-target",
            self.minimum_java_release,
        ]
        subprocess.check_call(javac + glob.glob("com/ubuntu/apport/*.java"))
        subprocess.check_call(
            ["jar", "cvf", "apport.jar"] + glob.glob("com/ubuntu/apport/*.class")
        )
        subprocess.check_call(javac + ["testsuite/crash.java"])
        subprocess.check_call(
            ["jar", "cvf", "crash.jar", "crash.class"], cwd="testsuite"
        )

        os.chdir(oldwd)


# pylint: disable-next=invalid-name
class install_java(Command):
    """Install Java components of Apport."""

    def __init__(self, dist: Distribution, **kwargs: dict[str, typing.Any]) -> None:
        super().__init__(dist, **kwargs)
        self.initialize_options()

    def initialize_options(self) -> None:
        """Set default values for all the options that this command supports."""
        self.install_dir: str | None = None

    def finalize_options(self) -> None:
        """Set final values for all the options that this command supports."""
        self.set_undefined_options("install_data", ("install_dir", "install_dir"))

    def _install_data_files(self, dst_path: str, src_files: list[pathlib.Path]) -> None:
        assert self.install_dir
        for src_file in src_files:
            target = pathlib.Path(self.install_dir) / dst_path / src_file.name
            self.mkpath(str(target.parent))
            self.copy_file(str(src_file), str(target), preserve_mode=False)

    def run(self) -> None:
        """Install the Java .jar files."""
        self._install_data_files("share/java", [pathlib.Path("java/apport.jar")])


@functools.cache
def has_java(unused_command: Command) -> bool:
    """Check if the Java compiler is available."""
    try:
        subprocess.run(["javac", "-version"], capture_output=True, check=True)
    except (OSError, subprocess.CalledProcessError):
        logging.getLogger(__name__).warning(
            "Java support: Java not available, not building Java crash handler"
        )
        return False
    return True


def register_java_sub_commands(
    build: type[Command], install: type[Command]
) -> dict[str, type[Command]]:
    """Plug the Java extension into setuptools.

    Return a dictionary with the added command classes which needs to
    be passed to the `setup` call as `cmdclass` parameter.
    """
    build.sub_commands.append(("build_java_subdir", has_java))
    install.sub_commands.append(("install_java", has_java))
    return {"build_java_subdir": build_java, "install_java": install_java}
